# (C) Copyright 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
RustPython 3.12.0 bytecode opcodes for version 0.40. There are other Rust 3.12 with different opcodes!
"""

#FIXME: this needs a lot of going over.

from typing import Dict, List, Optional, Tuple

import xdis.opcodes.opcode_3x.opcode_311 as opcode_311

# import xdis.opcodes.opcode_313 as opcode_313
from xdis.opcodes.base import finalize_opcodes, update_pj3
from xdis.opcodes.format.extended import extended_format_binary_op
from xdis.opcodes.opcode_3x.opcode_313 import opcode_arg_fmt313, opcode_extended_fmt313
from xdis.opcodes.opcode_rust.base import init_opdata_rust, make_opcodes
from xdis.version_info import PythonImplementation

_opcode = 0
version_tuple = (3, 12, 0)
python_implementation = PythonImplementation("RustPython")

# oppush[op] => number of stack entries pushed
oppush: List[int] = [0] * 256

# oppop[op] => number of stack entries popped
oppop: List[int] = [0] * 256

# opmap[opcode_name] => opcode_number
opmap: Dict[str, int] = {}

## pseudo opcodes (used in the compiler) mapped to the values
## they can become in the actual code.
_pseudo_ops = {}

_nb_ops = [
    ("NB_ADD", "+"),
    ("NB_AND", "&"),
    ("NB_FLOOR_DIVIDE", "//"),
    ("NB_LSHIFT", "<<"),
    ("NB_MATRIX_MULTIPLY", "@"),
    ("NB_MULTIPLY", "*"),
    ("NB_REMAINDER", "%"),
    ("NB_OR", "|"),
    ("NB_POWER", "**"),
    ("NB_RSHIFT", ">>"),
    ("NB_SUBTRACT", "-"),
    ("NB_TRUE_DIVIDE", "/"),
    ("NB_XOR", "^"),
    ("NB_INPLACE_ADD", "+="),
    ("NB_INPLACE_AND", "&="),
    ("NB_INPLACE_FLOOR_DIVIDE", "//="),
    ("NB_INPLACE_LSHIFT", "<<="),
    ("NB_INPLACE_MATRIX_MULTIPLY", "@="),
    ("NB_INPLACE_MULTIPLY", "*="),
    ("NB_INPLACE_REMAINDER", "%="),
    ("NB_INPLACE_OR", "|="),
    ("NB_INPLACE_POWER", "**="),
    ("NB_INPLACE_RSHIFT", ">>="),
    ("NB_INPLACE_SUBTRACT", "-="),
    ("NB_INPLACE_TRUE_DIVIDE", "/="),
    ("NB_INPLACE_XOR", "^="),
]

hasexc = []

loc = locals()

init_opdata_rust(loc, from_mod=None, version_tuple=version_tuple)

loc["opname"].extend([f"<{i}>" for i in range(256, 267)])
loc["oppop"].extend([0] * 11)
loc["oppush"].extend([0] * 11)

hasarg = []
hasconst = []
hasname = []
hasjrel = []
hasjabs = []
haslocal = []
hascompare = []
hasfree = []
hasexc = []

oplists = [
    loc["hasarg"],
    loc["hasconst"],
    loc["hasname"],
    loc["hasjrel"],
    loc["hasjabs"],
    loc["haslocal"],
    loc["hascompare"],
    loc["hasfree"],
    loc["hasexc"],
]

# add new table "hasjump"
loc.update({"hasjump": []})
loc["hasjrel"] = loc["hasjump"]

make_opcodes(loc, 12897)
EXTENDED_ARG = 103

# fmt: on

# for i in range(105):
#     print(f"%3d {loc['opname'][i]}" % i)

findlinestarts = opcode_311.findlinestarts

def extended_format_BINARY_OP(opc, instructions) -> Tuple[str, Optional[int]]:
    opname = _nb_ops[instructions[0].argval][1]
    if opname == "%":
        opname = "%%"
    elif opname == "%=":
        opname = "%%="
    return extended_format_binary_op(opc, instructions, f"%s {opname} %s")


opcode_extended_fmt313rust = {}
opcode_arg_fmt = opcode_arg_fmt13rust = {}

### update arg formatting
opcode_extended_fmt = opcode_extended_fmt312rust = {
    **opcode_extended_fmt313,
    **{
        "BINARY_OP": extended_format_BINARY_OP,
    },
}


update_pj3(globals(), loc, is_rust=True)
finalize_opcodes(loc)



# from xdis.bytecode import get_optype

from xdis.bytecode import Bytecode
from xdis.cross_dis import get_code_object
from xdis.instruction import Instruction
from xdis.lineoffsets_graal import find_linestarts_graal
from xdis.opcodes.opcode_graal.base_graal import (
    BINARY_OPS,
    UNARY_OPS,
    collection_to_str,
    get_optype_graal,
)


def get_instructions_bytes_graal(
    code_object,
    opc,
):
    """
    Iterate over the instructions in a bytecode string.

    Generates a sequence of Instruction namedtuples giving the details of each
    opcode.  Additional information about the code's runtime environment
    e.g., variable names, constants, can be specified using optional
    arguments.
    """
    bytecode: bytes = code_object.co_code
    constants: tuple = code_object.co_consts
    names: tuple = code_object.co_names
    varnames: tuple = code_object.co_varnames
    cells: tuple = code_object.co_cellvars
    freevars: tuple = code_object.co_freevars

    i = 0
    n = len(bytecode)

    labels = opc.findlabels(bytecode, opc)
    linestarts = find_linestarts_graal(code_object, opc, dup_lines=True)

    extended_arg_count = 0
    while i < n:
        opcode = bytecode[i]
        opname = opc.opname[opcode]
        optype = get_optype_graal(opcode, opc)
        offset = i
        starts_line = linestarts.get(offset, None)

        arg_count = opc.arg_counts[opcode]
        is_jump_target = i in labels

        # print(
        #     f"offset: {offset} {hex(opcode)} {opname} arg_count: {arg_count}, optype: {optype}"
        # )

        i += 1

        arg = -1
        argval = None
        argrepr = ""

        following_args = []
        has_arg = arg_count == 0

        if has_arg:
            argrepr = ""
        else:
            arg = bytecode[i]
            i += 1
            if arg_count > 1:
                following_args = []
                for j in range(arg_count - 1):
                    following_args.append(bytecode[i + j])
                    i += 1
                argrepr = str(arg)

        while True:
            if opcode == opc.opmap["EXTENDED_ARG"]:
                argrepr = ""
                break
            elif opcode in (opc.opmap["LOAD_BYTE_O"], opc.opmap["LOAD_BYTE_I"]):
                argrepr = str(arg)
                argval = arg
                break
            elif optype == "const":
                arg = arg
                argval = constants[arg]
                argrepr = str(constants[arg])
            elif opcode == opc.opmap["MAKE_FUNCTION"]:
                if following_args:
                    argrepr = str(following_args[0])
                argval = constants[arg]
                argrepr = argval.co_name
                break

            elif opcode in (opc.opmap["LOAD_INT"], opc.opmap["LOAD_LONG"]):
                argrepr = Objects.toString(primitiveConstants[arg])
                break
            elif opcode == opc.opmap["LOAD_DOUBLE"]:
                argrepr = Objects.toString(
                    Double.longBitsToDouble(primitiveConstants[arg])
                )
                break
            elif opcode == opc.opmap["LOAD_COMPLEX"]:
                real, imaginary = constants[arg]
                if imaginary == 0.0:
                    argval = imaginary
                else:
                    argval = complex(real, imaginary)
                argrepr = str(argval)
                break
            elif opcode in (
                opc.opmap["LOAD_CLOSURE"],
                opc.opmap["LOAD_DEREF"],
                opc.opmap["STORE_DEREF"],
                opc.opmap["DELETE_DEREF"],
            ):
                if arg >= len(cells):
                    argrepr = freevars[arg - len(cells)]
                else:
                    if 0 < arg < len(cells[arg]):
                        argrepr = cells[arg]
                    else:
                        argrepr = "???"

                break
            elif opcode in (
                opc.opmap["LOAD_FAST"],
                opc.opmap["STORE_FAST"],
                opc.opmap["DELETE_FAST"],
            ):
                argval = argrepr = varnames[arg]
                break

            elif optype == "name":
                argval = arg
                if 0 <= arg < len(names):
                    argrepr = names[arg]
                else:
                    argrepr = "??"
                break
            elif opcode == opc.opmap["FORMAT_VALUE"]:
                argval = arg
                kind = arg & 0x3
                if kind ==0x1:
                    argrepr = "STR"
                    break
                elif kind == 0x2:
                    argrepr = "REPR"
                    break
                elif kind == 0x3:
                    argrepr = "ASCII"
                    break
                elif kind == 0:
                    argrepr = "NONE"
                    break

                if (arg & 0x4) == 0x4:
                    argrepr += " + SPEC"
                    break

            elif opcode == opc.opmap["CALL_METHOD"]:
                argrepr = str(arg)
                break

            elif opcode == "unary":
                argval = arg
                argrepr = UNARY_OPS.get(arg, "??")
                break

            elif optype == "binary":
                argval = arg
                argrepr = BINARY_OPS.get(arg, "??")
                break

            elif optype == "collection":
                argval = arg
                argrepr = collection_to_str(arg)
                break
            elif opcode == opc.opmap["UNPACK_EX"]:
                argrepr = "%d, %d" % (arg, int(following_args[0]))
                break
            elif optype == "jrel":
                # fields.computeIfAbsent(offset + arg, k -> new String[DISASSEMBLY_NUM_COLUMNS])[1] = ">>"
                arg = arg
                if opcode == opc.opmap["JUMP_BACKWARD"]:
                    argval = offset - arg
                else:
                    argval = offset + arg
                argrepr = "to %d" % argval
                break
            else:
                pass
                # if opcode.quickens:
                #     opcode = opcode.quickens
                #     continue

            if opcode == opc.opmap["EXTENDED_ARG"]:
                arg <<= 8
            else:
                arg = 0
            break

        if arg_count == 0:
            arg = None

        inst_size = (arg_count + 1) + (extended_arg_count * 2)
        start_offset = offset if opc.oppop[opcode] == 0 else None

        # for (int i = 0 i < exceptionHandlerRanges.length; i += 4) {
        #     int start = exceptionHandlerRanges[i];
        #     int stop = exceptionHandlerRanges[i + 1];
        #     int handler = exceptionHandlerRanges[i + 2];
        #     int stackAtHandler = exceptionHandlerRanges[i + 3];
        #     String[] field = fields.get(handler);
        #     assert field != null;
        #     String handlerStr = String.format("exc handler %d - %d; stack: %d", start, stop, stackAtHandler);
        #     if (field[6] == null) {
        #         field[6] = handlerStr;
        #     } else {
        #         field[6] += " | " + handlerStr;
        #     }
        # }

        # for (i = 0; i < bytecode.length; i++) {
        #     String[] field = fields.get(i);
        #     if (field != null) {
        #         field[5] = field[5] == null ? "" : String.format("(%s)", field[5]);
        #         field[6] = field[6] == null ? "" : String.format("(%s)", field[6]);
        #         field[7] = "";
        #         if (outputCanQuicken != null && (outputCanQuicken[i] != 0 || generalizeInputsMap[i] != null)) {
        #             StringBuilder quickenSb = new StringBuilder();
        #             if (outputCanQuicken[i] != 0) {
        #                 quickenSb.append("can quicken");
        #             }
        #             if (generalizeInputsMap[i] != null) {
        #                 if (quickenSb.length() > 0) {
        #                     quickenSb.append(", ");
        #                 }
        #                 quickenSb.append("generalizes: ");
        #                 for (int j = 0; i < generalizeInputsMap[i].length; i++) {
        #                     if (j > 0) {
        #                         quickenSb.append(", ");
        #                     }
        #                     quickenSb.append(generalizeInputsMap[i][j]);
        #                 }
        #             }
        #             field[7] = quickenSb.toString();
        #         }
        #         formatted = "%-8s %2s %4s %-32s %-3s   %-32s %s %s" % field;
        #         sb.append(formatted.strip());
        #         sb.append('\n');
        #     }
        # }

        yield Instruction(
            is_jump_target=is_jump_target,
            starts_line= starts_line,
            offset=offset,
            opname=opname,
            opcode=opcode,
            has_arg=has_arg,
            arg=arg,
            argval=argval,
            argrepr=argrepr,
            tos_str=None,
            positions=None,
            optype=optype,
            inst_size=inst_size,
            has_extended_arg=extended_arg_count != 0,
            fallthrough=None,
            start_offset=start_offset,
        )


class Bytecode_Graal(Bytecode):
    """Bytecode operations involving a Python code object.

    Instantiate this with a function, method, string of code, or a code object
    (as returned by compile()).

    Iterating over these yields the bytecode operations as Instruction instances.
    """

    def __init__(self, x, opc, first_line=None, current_offset=None) -> None:
        self.codeobj = co = get_code_object(x)
        self._line_offset = 0
        self._cell_names = tuple()
        if first_line is None:
            self.first_line = co.co_firstlineno
        else:
            self.first_line = first_line
            self._line_offset = first_line - co.co_firstlineno
        pass

        self._linestarts = find_linestarts_graal(co, opc, dup_lines=True)
        self._original_object = x
        self.opc = opc
        self.opnames = opc.opname
        self.current_offset = current_offset

        if opc.version_tuple >= (3, 11) and hasattr(co, "co_exceptiontable"):
            self.exception_entries = None
            # self.exception_entries = parse_exception_table(co.co_exceptiontable)
        else:
            self.exception_entries = None

    def __iter__(self):
        co = self.codeobj
        return get_instructions_bytes_graal(
            co.co_code,
            self.opc,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._original_object!r})"

    def get_instructions(self, x):
        """Iterator for the opcodes in methods, functions or code

        Generates a series of Instruction named tuples giving the details of
        each operation in the supplied code.

        If *first_line* is not None, it indicates the line number that should
        be reported for the first source line in the disassembled code.
        Otherwise, the source line information (if any) is taken directly from
        the disassembled code object.
        """
        code_object = get_code_object(x)
        return get_instructions_bytes_graal(code_object, self.opc)



# (C) Copyright 2016-2017, 2020, 2023, 2025 by Rocky Bernstein
"""
CPython 3.2 bytecode opcodes

This is like Python 3.2's opcode.py with some classification
of stack usage and information for formatting instructions.
"""

import xdis.opcodes.opcode_3x.opcode_33 as opcode_33
from xdis.opcodes.base import (
    cpython_implementation,
    def_op,
    finalize_opcodes,
    init_opdata,
    rm_op,
    update_pj3,
)
from xdis.opcodes.opcode_3x.opcode_33 import opcode_arg_fmt33, opcode_extended_fmt33

# FIXME: can we DRY this even more?

python_implementation = cpython_implementation
version_tuple = (3, 2)

loc = locals()

init_opdata(loc, opcode_33, version_tuple)

# fmt: off
#              OP NAME               OPCODE POP PUSH
#--------------------------------------------------
def_op(loc,   "STOP_CODE",                0,  0,  0, fallthrough=False)
rm_op(loc, "YIELD_FROM", 72)

opcode_arg_fmt = opcode_arg_fmt32 = opcode_arg_fmt33
opcode_extended_fmt = opcode_extended_fmt32 = opcode_extended_fmt33

update_pj3(globals(), loc)
finalize_opcodes(loc)



# (C) Copyright 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from dataclasses import dataclass
from typing import Any, Dict, Tuple, Union

from xdis.codetype.code311 import Code311


@dataclass
class SourceLocation:
    # 1-based integer line number
    line_number: int
    # column offset in line: 1-based int (constructed from a zero-indexed stored value)
    column_offset: int


class Code313Rust(Code311):
    """Class for a RustPython 3.13 code object used when a Python
    interpreter is not RustPython 3.13 but working on RustPython 3.13 bytecode. It
    also functions as an object that can be used to build or write a
    Python3 code object, since we allow mutable structures.

    When done mutating, call method to_native().

    For convenience in generating code objects, fields like
    `co_consts`, co_names which are (immutable) tuples in the end-result can be stored
    instead as (mutable) lists. Likewise, the line number table `co_lnotab`
    can be stored as a simple list of offset, line_number tuples.

    """
    def __init__(
        self,
        co_argcount: int,
        co_posonlyargcount: int,
        co_kwonlyargcount: int,
        co_nlocals: int,
        co_stacksize: int,
        co_flags: int,
        co_code: bytes,
        co_consts: tuple,
        co_names: tuple[str],
        co_varnames: tuple[str],
        co_filename: str,
        co_name: str,
        co_qualname: str,
        co_firstlineno: int,
        co_linetable: bytes,
        co_freevars: tuple,
        co_cellvars: tuple,
        version_triple: Tuple[int, int, int],
        locations: tuple,
        co_exceptiontable = tuple(),
        collection_order: Dict[Union[set, frozenset, dict], Tuple[Any]] = {},
    ) -> None:
        self.co_argcount = co_argcount
        self.co_posonlyargcount = co_posonlyargcount
        self.co_kwonlyargcount = co_kwonlyargcount
        self.co_nlocals = co_nlocals
        self.co_stacksize = co_stacksize
        self.co_flags = co_flags
        self.co_code = co_code
        self.co_consts = co_consts
        self.co_names = co_names
        self.co_varnames = co_varnames
        self.co_filename = co_filename
        self.co_name = co_name
        self.co_firstlineno = co_firstlineno    # None if 0 in serialized form
        self.co_linetable = co_linetable
        self.co_qualname = co_qualname
        self.co_cellvars = co_cellvars
        self.co_freevars= co_freevars
        self.co_exceptiontable = co_exceptiontable
        self.co_collection_order = collection_order
        self.version_triple = version_triple
        self.locations = locations



# (C) Copyright 2023, 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Bytecode opcode modules with some classification
of stack usage and information for formatting instructions.
This covers information from the Python stdlib opcodes.py.
"""

from xdis.opcodes.opcode_1x import (
    opcode_10,
    opcode_11,
    opcode_12,
    opcode_13,
    opcode_14,
    opcode_15,
    opcode_16,
)
from xdis.opcodes.opcode_2x import (
    opcode_20,
    opcode_21,
    opcode_22,
    opcode_23,
    opcode_24,
    opcode_25,
    opcode_26,
    opcode_27,
)
from xdis.opcodes.opcode_3x import (
    opcode_30,
    opcode_31,
    opcode_32,
    opcode_33,
    opcode_34,
    opcode_35,
    opcode_36,
    opcode_37,
    opcode_38,
    opcode_39,
    opcode_310,
    opcode_311,
    opcode_312,
    opcode_313,
    opcode_314,
)
from xdis.opcodes.opcode_pypy import (
    opcode_26pypy,
    opcode_27pypy,
    opcode_32pypy,
    opcode_33pypy,
    opcode_35pypy,
    opcode_36pypy,
    opcode_37pypy,
    opcode_38pypy,
    opcode_39pypy,
    opcode_310pypy,
    opcode_311pypy,
)
from xdis.opcodes.opcode_rust import opcode_3531rust, opcode_12897rust, opcode_24481rust

# from xdis.opcodes.opcode_graal import (
#     opcode_310graal,
#     opcode_311graal,
#     opcode_312graal,
#     opcode_38graal,
#     )


__all__ = [
    "opcode_10",
    "opcode_11",
    "opcode_12",
    "opcode_13",
    "opcode_14",
    "opcode_15",
    "opcode_16",
    "opcode_20",
    "opcode_21",
    "opcode_22",
    "opcode_23",
    "opcode_24",
    "opcode_25",
    "opcode_26",
    "opcode_26pypy",
    "opcode_27",
    "opcode_27pypy",
    "opcode_30",
    "opcode_31",
    "opcode_310",
    "opcode_311",
    "opcode_312",
    "opcode_313",
    "opcode_314",
    "opcode_32",
    "opcode_33",
    "opcode_34",
    "opcode_35",
    "opcode_36",
    "opcode_37",
    "opcode_38",
    "opcode_39",
    "opcode_310pypy",
    "opcode_311pypy",
    "opcode_32pypy",
    "opcode_33pypy",
    "opcode_35pypy",
    "opcode_36pypy",
    "opcode_37pypy",
    "opcode_38pypy",
    "opcode_39pypy",
    # "opcode_310graal",
    # "opcode_311graal",
    # "opcode_312rust",
    "opcode_3531rust",
    "opcode_12897rust",
    "opcode_24481rust",
]



# (C) Copyright 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Common routines for entering and classifying rust opcodes. Inspired by,
limited by, and somewhat compatible with the corresponding
Python opcode.py structures
"""

from xdis.opcodes.base import (
    VARYING_STACK_INT,
    binary_op,
    call_op,
    compare_op,
    conditional_op,
    const_op,
    def_op,
    free_op,
    init_opdata,
    jabs_op,
    jrel_op,
    local_op,
    name_op,
    nargs_op,
    store_op,
    ternary_op,
    unary_op,
    varargs_op,
)

# RustPython compare operation enumerations are different from CPythons
# From bytecode.rs:
#    be intentional with bits so that we can do eval_ord with just a bitwise and
#    bits: | Equal | Greater | Less |

cmp_op = (
    "==",  # 0
    "<",  # 1
    ">",  # 2
    "!=",  # 3
    "==",  # 4
    ">",  # 5
    "<=",  # 5
    ">=",  # 6
)

test_op = (
    "in",  # 0
    "not-in",  # 1
    "is",  # 2
    "is-not",  # 3
    "exception-match",  # 4
)

def init_opdata_rust(loc, from_mod, version_tuple) -> None:
    """Sets up a number of the structures found in Python's
    opcode.py. Python opcode.py routines assign attributes to modules.
    In order to do this in a modular way here, the local dictionary
    for the module is passed.
    """
    init_opdata(loc, from_mod, version_tuple, False)
    loc["cmp_op"] = cmp_op

def inc_opcode(loc):
    loc["_opcode"] += 1

def binary_op_graal(loc: dict, name: str, pop: int = 2, push: int = 1) -> None:
    """
    Put opcode in the class of instructions that are binary operations.
    """
    binary_op(loc, name, loc["_opcode"], pop, push)
    inc_opcode(loc)


def call_op_graal(
    loc: dict, name: str, pop: int = -2, push: int = 1, fallthrough: bool=True
) -> None:
    """
    Put opcode in the class of instructions that perform calls.
    """
    call_op(loc, name, loc["_opcode"], pop, push, fallthrough)
    inc_opcode(loc)


def compare_op_graal(loc: dict, name: str, pop: int = 2, push: int = 1) -> None:
    compare_op(loc, name, loc["_opcode"], pop, push)
    inc_opcode(loc)


def conditional_op_graal(loc: dict, name: str) -> None:
    conditional_op(loc, name, loc["_opcode"])
    inc_opcode(loc)


def const_op_graal(loc: dict, name: str, pop: int = 0, push: int = 1) -> None:
    const_op(loc, name, loc["_opcode"], pop, push)
    inc_opcode(loc)


def def_op_graal(
    loc: dict,
    op_name: str,
    pop: int = -2,
    push: int = -2,
    fallthrough: bool = True,
) -> None:
    def_op(loc, op_name, loc["_opcode"], pop, push, fallthrough)
    inc_opcode(loc)


def free_op_graal(loc: dict, op_name: str, pop: int = 0, push: int = 1) -> None:
    free_op(loc, op_name, loc["_opcode"], pop, push)
    inc_opcode(loc)


def jabs_op_graal(
    loc: dict,
    op_name: str,
    pop: int = 0,
    push: int = 0,
    conditional: bool = False,
    fallthrough: bool = True,
) -> None:
    """
    Put opcode in the class of instructions that can perform an absolute jump.
    """
    jabs_op(loc, op_name, loc["_opcode"], pop, push, conditional, fallthrough)
    inc_opcode(loc)


def jrel_op_graal(loc, op_name: str, pop: int=0, push: int=0, conditional=False, fallthrough=True) -> None:
    """
    Put opcode in the class of instructions that can perform a relative jump.
    """
    jrel_op(loc, op_name, loc["_opcode"], pop, push, conditional, fallthrough)
    inc_opcode(loc)


def local_op_graal(loc, op_name, pop=0, push=1) -> None:
    local_op(loc, op_name, loc["_opcode"], pop, push)
    inc_opcode(loc)


def name_op_graal(loc: dict, op_name, pop=-2, push=-2) -> None:
    """
    Put opcode in the class of instructions that index into the "name" table.
    """
    name_op(loc, op_name, loc["_opcode"], pop, push)
    inc_opcode(loc)


def nargs_op_graal(
    loc, op_name: str, pop: int = -2, push: int = -1, fallthrough=True
) -> None:
    """
    Put opcode in the class of instructions that have a variable number of (or *n*) arguments
    """
    nargs_op(loc, op_name, loc["_opcode"], pop, push, fallthrough)
    inc_opcode(loc)


def store_op_graal(loc: dict, op_name: str, pop=0, push=1, is_type="def") -> None:
    store_op(loc, op_name, loc["_opcode"], pop, push, is_type)
    inc_opcode(loc)


def ternary_op_graal(loc: dict, op_name: str, pop: int = 3, push: int = 1) -> None:
    """
    Put opcode in the class of instructions that are ternary operations.
    """
    ternary_op(loc, op_name, loc["_opcode"], pop, push)
    inc_opcode(loc)


def unary_op_graal(loc, op_name: str, pop: int=1, push: int=1) -> None:
    unary_op(loc, op_name, loc["_opcode"], pop, push)
    inc_opcode(loc)


def varargs_op_graal(loc, op_name, pop: int=-1, push: int=1) -> None:
    varargs_op(loc, op_name, loc["_opcode"], pop, push)
    inc_opcode(loc)



# See RustPython/compiler/core/src/bytecode.rs Instruction
# fmt: off

def make_opcodes(loc: dict, magic_int: int):
    #            OP NAME                        POP PUSH
    #----------------------------------------------------------
    def_op_graal(loc,    "NOP",             0, 0)
    name_op_graal(loc,   "IMPORT_NAME")
    def_op_graal(loc,    "IMPORT_NAMELESS",     0, 1)
    name_op_graal(loc,   "IMPORT_FROM")

    local_op_graal(loc,  "LOAD_FAST")
    name_op_graal(loc,   "LOAD_NAME")
    name_op_graal(loc,   "LOAD_GLOBAL")
    free_op_graal(loc,   "LOAD_DEREF")

    def_op_graal(loc,    "LOAD_CLASS_DEREF")

    store_op_graal(loc,  "STORE_FAST",         1, 0, is_type="local")  # Local variable
    store_op_graal(loc,  "STORE_LOCAL",        0)
    name_op_graal(loc,   "STORE_GLOBAL")                                    # ""
    store_op_graal(loc,  "STORE_DEREF",        1, 0, is_type="free")

    local_op_graal(loc,  "DELETE_FAST",        0, 0) # Local variable number is in operand
    local_op_graal(loc,  "DELETE_LOCAL",       0, 0)
    name_op_graal(loc,   "DELETE_GLOBAL")            # ""
    free_op_graal(loc,   "DELETE_DEREF",       0, 0)
    free_op_graal(loc,   "LOAD_CLOSURE")
    binary_op_graal(loc, "BINARY_SUBSCR")

    def_op_graal(loc,    "STORE_SUBSCR",       0, 0)
    def_op_graal(loc,    "DELETE_SUBSCR",      0, 0)
    name_op_graal(loc,   "STORE_ATTR")               # Index in name list
    name_op_graal(loc,   "DELETE_ATTR")              # ""

    # Operand is in const list
    const_op_graal(loc,  "LOAD_CONST",         0, 1)
    unary_op_graal(loc,  "UNARY_OP",           1, 1)
    binary_op_graal(loc, "BINARY_OP")
    binary_op_graal(loc, "BINARY_OP_INPLACE")
    if magic_int in (12897,):
        name_op_graal(loc,   "LOAD_ATTR")
    compare_op_graal(loc, "TEST_OP",           2, 1)  # test operator
    if magic_int in (24881,):
        def_op_graal(loc,     "COPY")
        binary_op_graal(loc, "BINARY_SUBCR")
    compare_op_graal(loc, "COMPARE_OP",        2, 1)  # Comparison operator
    if magic_int in (24881,):
        def_op_graal(loc,     "SWAP",          1, 1)
    def_op_graal(loc,     "POP_TOP",           1, 0)
    if magic_int in (24881,):
        def_op_graal(loc,     "TO_BOOL",       1, 1)
        def_op_graal(loc,     "Unknown1")
    def_op_graal(loc,     "ROT_TWO",           2, 2)
    def_op_graal(loc,     "ROT_THREE",         3, 3)
    def_op_graal(loc,     "DUP_TOP",           0, 1)
    def_op_graal(loc,     "Duplicate2")
    def_op_graal(loc,     "GET_ITER")

    if magic_int in (24881,):
        def_op_graal(loc,     "GetLen")
        call_op_graal(loc,     "CallIntrinsic1")
        call_op_graal(loc,     "CallIntrinsic2")
    def_op_graal(loc,     "CONTINUE_LOOP")
    def_op_graal(loc,     "BREAK_LOOP")

    # In contrast to CPython and others Rust does not seem to use relative jumps,
    # but only absolute jumps.
    jabs_op_graal(loc, "JUMP")    # Number of words to skip

    jabs_op_graal(loc, "POP_JUMP_IF_TRUE")
    jabs_op_graal(loc, "POP_JUMP_IF_FALSE")
    jabs_op_graal(loc, "JUMP_IF_TRUE_OR_POP") # Number of words to skip
    jabs_op_graal(loc, "JUMP_IF_FALSE_OR_POP") # "

    nargs_op_graal(loc, "MAKE_FUNCTION", VARYING_STACK_INT)    # Flags

    if magic_int in (24881,):
        def_op_graal(loc, "SET_FUNCTION_ATTR")

    call_op_graal(loc, "CALL_FUNCTION")
    call_op_graal(loc, "CALL_FUNCTION_KW")
    call_op_graal(loc, "CALL_FUNCTION_EX")  # Flags
    def_op_graal(loc, "LOAD_METHOD")
    call_op_graal(loc, "CALL_METHOD")
    call_op_graal(loc, "CALL_METHOD_KW")
    call_op_graal(loc, "CALL_METHOD_EX")

    jabs_op_graal(loc, "FOR_ITER")
    def_op_graal(loc, "RETURN_VALUE")
    const_op_graal(loc, "RETURN_CONST")
    def_op_graal(loc, "YIELD")
    def_op_graal(loc, "YIELD_FROM")
    if magic_int in (24881,):
        def_op_graal(loc, "RESUME", 0, 0)

    def_op_graal(loc, "SETUP_ANNOTATIONS")
    jabs_op_graal(loc, "SETUP_LOOP", 0,  0, conditional=True)
    def_op_graal(loc, "SETUP_FINALLY", 0, 1)

    if magic_int in (12897,):
        def_op_graal(loc,  "EnterFinally")
        def_op_graal(loc,  "EndFinally")

    if magic_int in (24881,):
        def_op_graal(loc, "GET_YIELD_FROM_ITER")
    def_op_graal(loc,  "SETUP_EXCEPT")
    def_op_graal(loc, "SETUP_WITH")
    def_op_graal(loc,  "SETUP_CLEANUP")
    def_op_graal(loc,  "WithCleanupStart")
    if magic_int in (24881,):
        def_op_graal(loc,  "WithCleanupFinish")

    def_op_graal(loc, "POP_BLOCK")
    def_op_graal(loc, "RAISE")

    if magic_int in (24881,):
        def_op_graal(loc, "LIST_TO_TUPLE")

    if magic_int in (12897,):
        varargs_op_graal(loc, "BUILD_STRING", 1, VARYING_STACK_INT)      # Number of tuple items

    varargs_op_graal(loc, "BUILD_TUPLE", 1, VARYING_STACK_INT)      # Number of tuple items

    if magic_int in (12897,):
        varargs_op_graal(loc, "BUILD_TUPLE_UNPACK", 1, VARYING_STACK_INT)      # Number of tuple items

    if magic_int in (24881,):
        varargs_op_graal(loc, "BUILD_TUPLE_FROM_TUPLES", 1, VARYING_STACK_INT)      # Number of tuple items
        varargs_op_graal(loc, "BUILD_TUPLE_FROM_ITER", 1, VARYING_STACK_INT)      # Number of tuple items
    varargs_op_graal(loc, "BUILD_LIST", VARYING_STACK_INT)       # Number of list items

    if magic_int in (12897,):
        varargs_op_graal(loc, "BUILD_LIST_UNPACK", 1, VARYING_STACK_INT)       # Number of list items
    if magic_int in (24881,):
        varargs_op_graal(loc, "BUILD_LIST_FROM_TUPLES", 1, VARYING_STACK_INT)       # Number of list items
    varargs_op_graal(loc, "BUILD_SET", 1, VARYING_STACK_INT)        # Number of set items
    if magic_int in (12897,):
        varargs_op_graal(loc, "BUILD_SET", 1, VARYING_STACK_INT)        # Number of set items
    if magic_int in (24881,):
        varargs_op_graal(loc, "BUILD_SET_FROM_TUPLES", 1, VARYING_STACK_INT)       # Number of list items
    varargs_op_graal(loc, "BUILD_MAP", 1, VARYING_STACK_INT)        # Number of dict entries
    varargs_op_graal(loc, "BUILD_MAP_FOR_CALL", 1, VARYING_STACK_INT)        # Number of dict entries
    varargs_op_graal(loc, "DICT_UPDATE")        # Number of dict entries
    def_op_graal(loc, "BUILD_SLICE")
    def_op_graal(loc, "LIST_APPEND")
    def_op_graal(loc, "SET_ADD")
    def_op_graal(loc, "MAP_ADD")
    def_op_graal(loc, "PRINT_EXPR")
    def_op_graal(loc, "LOAD_BUILD_CLASS")
    varargs_op_graal(loc, "UNPACK_SEQUENCE", 1, VARYING_STACK_INT)   # Number of tuple items
    def_op_graal(loc, "UNPACK_EX", VARYING_STACK_INT, VARYING_STACK_INT)
    def_op_graal(loc, "FORMAT_VALUE", VARYING_STACK_INT, VARYING_STACK_INT)
    def_op_graal(loc, "POP_EXCEPT")
    def_op_graal(loc, "REVERSE")
    def_op_graal(loc, "GET_AWAITABLE")

    if magic_int in (12897,):
        def_op_graal(loc, "BeforeAsyncWith")
        def_op_graal(loc, "SetupAsyncWith")
        def_op_graal(loc, "GET_AITER", 0, 1)
        def_op_graal(loc, "GET_ANEXT")
        def_op_graal(loc, "EndAsyncFor")

    if magic_int in (24881,):
        def_op_graal(loc, "END_ASYNC_FOR")
        def_op_graal(loc, "MATCH_MAPPING")
        def_op_graal(loc, "MATCH_SEQUENCE")
        def_op_graal(loc, "MATCH_CLASS")

    def_op_graal(loc, "EXTENDED_ARG")

    if magic_int in (12897,):
        def_op_graal(loc, "TypeVar")
        def_op_graal(loc, "TypeVarWithBound")
        def_op_graal(loc, "TypeVarWithConstraint")
        def_op_graal(loc, "TypeAlias")



# (C) 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import types
from copy import deepcopy
from types import CodeType
from typing import Any, Dict, Set, Tuple

from xdis.codetype.code38 import Code38, Code38FieldTypes
from xdis.version_info import PYTHON_VERSION_TRIPLE, version_tuple_to_str

# Note: order is the positional order given in the Python docs for
# 3.11 types.Codetype.
# "posonlyargcount" is not used, but it is in other Python versions, so it
# has to be included since this structure is used as the Union type
# for all code types.
Code38GraalFieldNames = """
        co_argcount
        co_cellvars
        co_code
        co_consts
        co_exceptiontable
        co_filename
        co_firstlineno
        co_flags
        co_freevars
        co_kwonlyargcount
        co_lnotab
        co_name
        co_names
        co_nlocals
        co_posonlyargcount
        co_stacksize
        co_varnames
        condition_profileCount
        endColumn
        endLine
        exception_handler_ranges
        generalizeInputsMap
        generalizeVarsMap
        outputCanQuicken
        primitiveConstants
        srcOffsetTable
        startColumn
        startLine
        variableShouldUnbox
"""

Code38GraalFieldTypes = deepcopy(Code38FieldTypes)

class Code38Graal(Code38):
    """Class for a Python 3.11+ code object used when a Python interpreter less than 3.11 is
    working on Python 3.11 bytecode. It also functions as an object that can be used
    to build or write a Python3 code object, since we allow mutable structures.

    When done mutating, call method to_native().

    For convenience in generating code objects, fields like
    `co_consts`, co_names which are (immutable) tuples in the end-result can be stored
    instead as (mutable) lists. Likewise, the line number table `co_linetable`
    can be stored as a simple list of offset, line_number tuples.
    """

    def __init__(
        self,
        co_argcount: int,
        co_posonlyargcount: int,
        co_kwonlyargcount: int,
        co_nlocals: int,
        co_stacksize: int,
        co_flags: int,
        co_consts: tuple,
        co_code,
        co_names: tuple,
        co_varnames: tuple,
        co_filename,
        co_name,
        co_firstlineno,
        co_lnotab,
        co_freevars,
        co_cellvars,
        reference_objects: Set[Any] = set(),
        version_triple: Tuple[int, int, int] = (0, 0, 0),
        other_fields: Dict[str, Any] = {},
    ) -> None:
        # Keyword argument parameters in the call below is more robust.
        # Since things change around, robustness is good.
        super().__init__(
            co_argcount=co_argcount,
            co_cellvars=co_cellvars,
            co_code=co_code,
            co_consts=co_consts,
            co_filename=co_filename,
            co_firstlineno=co_firstlineno,
            co_flags=co_flags,
            co_freevars=co_freevars,
            co_kwonlyargcount=co_kwonlyargcount,
            co_lnotab=co_lnotab,
            co_name=co_name,
            co_names=co_names,
            co_nlocals=co_nlocals,
            co_posonlyargcount=co_posonlyargcount,
            co_stacksize=co_stacksize,
            co_varnames=co_varnames,
            reference_objects=reference_objects,
            version_triple=version_triple,
        )

        for field_name, value in other_fields.items():
            setattr(self, field_name, value)

        self.fieldtypes = Code38GraalFieldTypes
        if type(self) is Code38Graal:
            self.check()

    def to_native(self) -> CodeType:
        if not (PYTHON_VERSION_TRIPLE >= (3, 11)):
            raise TypeError(
                "Python Interpreter needs to be in 3.11 or greater; is %s"
                % version_tuple_to_str()
            )

        code = deepcopy(self)
        code.freeze()
        try:
            code.check()
        except AssertionError as e:
            raise TypeError(e)

        return types.CodeType(
            code.co_argcount,
            code.co_posonlyargcount,
            code.co_kwonlyargcount,
            code.co_nlocals,
            code.co_stacksize,
            code.co_flags,
            code.co_code,
            code.co_consts,
            code.co_names,
            code.co_varnames,
            code.co_filename,
            code.co_name,
            code.co_firstlineno,
            code.co_lnotab,
            code.co_freevars,
            code.co_cellvars,
        )



"""

This module is an adaptation of the GraalPython's Java SourceMap helper

The original assumes Truffle's Source/SourceSection APIs.

   * The get_source_section(source, ...) method expects a "source" object that implements:
       - has_characters() -> bool
      - get_line_length(line: int) -> int
       - create_unavailable_section() -> any
       - create_section(start_line, start_column, end_line, end_column) -> any
   * The SourceMap constructor accepts an optional op_length_fn parameter (callable)
     that takes a single byte value (0..255) and returns the length (int) of the
     instruction at that BCI. If omitted, all instructions are assumed to have length 1.

 - Byte encoding/decoding handles Java's signed byte semantics (values -128..127).
 - This is intended to reproduce logic and encoding used in the Java original. It is
   not tied to any Truffle internals and can be adapted to a concrete environment.

See https://github.com/oracle/graalpython/graalpython/com.oracle.graal.python/src/com/oracle/graal/python/compiler/SourceMap.java
"""

from typing import Dict, List, Tuple

# Signed byte constants (as in Java code)
EXTENDED_NUM = -128
NEXT_LINE = -127
NEXT_LINES = -126
MIN_NUM = -125
MULTIPLIER_NEGATIVE = MIN_NUM + 1  # -124
MAX_NUM = 127
MULTIPLIER_POSITIVE = MAX_NUM  # 127


def _to_signed(byte_value: int) -> int:
    """Convert 0..255 byte to signed -128..127 value like Java byte."""
    return byte_value if byte_value < 128 else byte_value - 256


class SourceMap:
    """
    Encapsulates encoding/decoding of source line/column information for GraalPython bytecode.

    Constructor:
      SourceMap(code: bytes, src_table: bytes, start_line: int, start_column: int,
                op_length_fn: Optional[Callable[[int], int]] = None)

    - code: bytes of bytecode (used to determine map size and iterate instruction BCIs)
    - src_table: the encoded source table bytes produced by Builder
    - start_line, start_column: initial line/column used to start decoding
    - op_length_fn: optional callable taking an opcode byte (0..255) and returning
      the instruction length in bytes. If None, instructions are assumed length 1.
    """

    def __init__(
        self,
        code_object,
        opc,
    ):
        # use Python lists (mutable) to store maps; they match length of code

        bytecode: bytes = code_object.co_code
        start_column: int  = code_object.startColumn
        start_line: int  = code_object.startLine
        # cells: tuple = code_object.co_cellvars
        # freevars: tuple = code_object.co_freevars
        arg_counts: Dict[int, int] = opc.arg_counts


        n = len(bytecode)
        self.startLineMap: List[int] = [0] * n
        self.endLineMap: List[int] = [0] * n
        self.startColumnMap: List[int] = [0] * n
        self.endColumnMap: List[int] = [0] * n
        self.source_table: bytes = code_object.srcOffsetTable
        self.source_table_len: int = len(self.source_table)

        # op_length_fn determines instruction size; default to 1

        self.source_table_pos = 0

        self.next_column = start_column
        self.next_line = start_line
        offset = 0
        # print(f"XXX0 len: {n}") # debug
        while offset < n:

            # code[offset] is an int 0..255 in Python 3 when indexing bytes
            # print(f"offset: {offset}") # debug
            op_byte = bytecode[offset]
            # print(f"{opc.opname[op_byte]}") # debug
            op_len = arg_counts[op_byte] + 1
            # print(f"op_len: {op_len}") # debug

            try:
                start_line, start_column = self._next_line_and_column()
                end_line, end_column = self._next_line_and_column()
            except EOFError:
                # FIXME: We made a mistake somewhere.
                break

            # fill maps for the bytes covered by this opcode
            for i in range(offset, min(offset + op_len, n)):
                self.startLineMap[i] = start_line
                self.startColumnMap[i] = start_column
                self.endLineMap[i] = end_line
                self.endColumnMap[i] = end_column
            offset += op_len

    def _next_line_and_column(self) -> Tuple[int, int]:
        """
        Get the (line, column) pair delta from self.source_table.

        Java code (which reads from a stream):
            stream.mark(1);
            byte value = (byte) stream.read();
            if (value == NEXT_LINE) { pair[0]++; pair[1] = 0; }
            else if (value == NEXT_LINES) { pair[0] += readNum(stream); pair[1] = 0; }
            else { stream.reset(); }
            pair[1] += readNum(stream);
        """
        if self.source_table_pos >= self.source_table_len:
            raise EOFError("Unexpected end of source table while reading line/column")

        old_pos = self.source_table_pos
        b = self.source_table[self.source_table_pos]
        self.source_table_pos += 1
        v = _to_signed(b)
        if v == NEXT_LINE:
            self.next_line += 1
            self.next_column = 0
        elif v == NEXT_LINES:
            self.next_line += self._get_num()
            self.next_column = 0
        else:
            # reset to before the byte we consumed
            self.source_table_pos = old_pos
        self.next_column += self._get_num()
        return self.next_line, self.next_column

    def _get_num(self) -> int:
        """
        Decode a signed (possibly extended) number from the stream.

        This Behavior mirrors Java's readNum using EXTENDED_NUM and multipliers,
        but that code reads from a stream insttead of self.source_table.
        """
        extensions = 0
        while True:
            if self.source_table_pos >= self.source_table_len:
                raise EOFError("Unexpected end of source table while reading line/column")

            b = self.source_table[self.source_table_pos]
            self.source_table_pos += 1
            val = _to_signed(b)
            if val == EXTENDED_NUM:
                extensions += 1
            elif val < 0:
                # negative single-byte value (signed)
                return extensions * MULTIPLIER_NEGATIVE + val
            else:
                # non-negative single-byte value
                return extensions * MULTIPLIER_POSITIVE + val

def find_linestarts_graal(code_object, opc, dup_lines: bool) -> dict:
    source_map = SourceMap(code_object, opc)
    bytecode: bytes = code_object.co_code
    i = 0
    n = len(bytecode)
    last_lineno = -1
    offset2line: Dict[int, int] = {}
    lines_seen = set()
    last_linemap_offset = len(source_map.startLineMap) - 1
    while i < n:
        opcode = bytecode[i]
        offset = i
        i += opc.arg_counts[opcode] + 1
        if offset >= last_linemap_offset:
            break
        line_number = source_map.startLineMap[offset]
        if line_number != last_lineno:
            if not dup_lines:
                if line_number in lines_seen:
                    continue
                else:
                    lines_seen.add(line_number)
            offset2line[offset] = line_number

        last_lineno = line_number

    return offset2line



# Copyright (c) 2015-2021, 2024-2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""RustPython version-independent Python object deserialization (unmarshal).

This is needed when the bytecode extracted is from
a different version than the currently-running Python.

When the running interpreter and the read-in bytecode are the same,
you can use Python's built-in ``marshal.loads()`` to produce a code
object.
"""

from typing import Any, Dict, List, Tuple, Union

from xdis.codetype.code312rust import Code312Rust
from xdis.codetype.code313rust import Code313Rust, SourceLocation
from xdis.magics import magic_int2tuple
from xdis.unmarshal import (
    TYPE_ASCII,
    TYPE_ASCII_INTERNED,
    TYPE_BINARY_COMPLEX,
    TYPE_BINARY_FLOAT,
    TYPE_CODE,
    TYPE_DICT,
    TYPE_ELLIPSIS,
    TYPE_FALSE,
    TYPE_FROZENSET,
    TYPE_INT,
    TYPE_INT64,
    TYPE_INTERNED,
    TYPE_LIST,
    TYPE_LONG,
    TYPE_NONE,
    TYPE_NULL,
    TYPE_REF,
    TYPE_SET,
    TYPE_SHORT_ASCII,
    TYPE_SHORT_ASCII_INTERNED,
    TYPE_SLICE,
    TYPE_SMALL_TUPLE,
    TYPE_STOPITER,
    TYPE_STRING,
    TYPE_STRINGREF,
    TYPE_TRUE,
    TYPE_TUPLE,
    TYPE_UNICODE,
    TYPE_UNKNOWN,
    VersionIndependentUnmarshaller,
    compat_str,
)
from xdis.version_info import version_tuple_to_str

# Bit set on marshalType if we should
# add obj to intern_objects.
# FLAG_REF is the marshal.c name
FLAG_REF = 0x80

TYPE_COMPLEX_RUSTPYTHON = "y"
TYPE_FLOAT_RUSTPYTHON = "g"

# The keys in the following dictionary are unmarshal codes, like "s",
# "c", "<", etc. The values of the dictionary are names of routines
# to call that do the data unmarshaling.
#
# Note: we could eliminate the parameters if this were all inside a
# class.  This might be good from an efficiency standpoint, and bad
# from a functional-programming standpoint. Pick your poison.
# EDIT: I'm choosing efficiency over functional programming.
UNMARSHAL_DISPATCH_TABLE = {
    TYPE_ASCII: "ASCII",
    TYPE_ASCII_INTERNED: "ASCII_interned",
    TYPE_BINARY_COMPLEX: "binary_complex",
    TYPE_BINARY_FLOAT: "binary_float",
    TYPE_COMPLEX_RUSTPYTHON: "complex",
    TYPE_CODE: "code_rust",
    TYPE_DICT: "dict",
    TYPE_ELLIPSIS: "Ellipsis",
    TYPE_FALSE: "False",
    TYPE_FLOAT_RUSTPYTHON: "float",
    TYPE_FROZENSET: "frozenset",
    TYPE_INT64: "int64",
    TYPE_INT: "bigint",
    TYPE_INTERNED: "interned",
    TYPE_LIST: "list",
    TYPE_LONG: "long",
    TYPE_NONE: "None",
    TYPE_NULL: "C_NULL",
    TYPE_REF: "object_reference",
    TYPE_SET: "set",
    TYPE_SHORT_ASCII: "short_ASCII",
    TYPE_SHORT_ASCII_INTERNED: "short_ASCII_interned",
    TYPE_SLICE: "slice",
    TYPE_SMALL_TUPLE: "small_tuple",
    TYPE_STOPITER: "stopIteration",
    TYPE_STRING: "string",
    TYPE_STRINGREF: "python2_string_reference",
    TYPE_TRUE: "True",
    TYPE_TUPLE: "tuple",
    TYPE_UNICODE: "unicode",
    TYPE_UNKNOWN: "unknown",
}

class MarshalError(Exception):
    pass

class VersionIndependentUnmarshallerRust(VersionIndependentUnmarshaller):
    def __init__(self, fp, magic_int, bytes_for_s, code_objects={}) -> None:
        """
        Marshal versions:
            4: [3.4a3, 3.13) (self.magic_int: 3280 onwards)
            5: [3.14, current) (self.magic_int: circa 3608)

        In Python 3, a ``bytes`` type is used for strings.
        """
        self.fp = fp
        self.magic_int = magic_int
        self.code_objects = code_objects

        # Save a list of offsets in the bytecode file where code
        # objects starts.
        self.code_to_file_offsets = {}

        # It is helpful to save the order in sets, frozensets and dictionary keys,
        # so that on writing a bytecode file we can duplicate this order.
        self.collection_order: Dict[Union[set, frozenset, dict], Tuple[Any]] = {}

        self.bytes_for_s = bytes_for_s
        self.version_triple = magic_int2tuple(self.magic_int)
        if self.version_triple >= (3, 14):
            self.marshal_version = 5
        elif (3, 14) > self.version_triple >= (3, 4):
            if self.magic_int in (12897,):
                self.marshal_version = 3
            else:
                self.marshal_version = 4
        else:
            assert False, f"version {version_tuple_to_str(self.version.triple)} is not a graal version"

        self.intern_strings = []
        self.intern_objects = []
        self.is_graal = False
        self.is_pypy = False
        self.is_rust = True

        self.UNMARSHAL_DISPATCH_TABLE = UNMARSHAL_DISPATCH_TABLE

    def t_code_rust(self, save_ref, bytes_for_s: bool = False) -> Code313Rust:
        # read instructions
        instr_count = self.read_int32()
        co_code = self.read_slice(instr_count * 2)

        # instructions = [int(co_code[i]) for i in range(len(co_code))] # debug
        co_lnotab = {}

        # read locations
        loc_count = self.read_int32()
        locations: List[SourceLocation] = []
        last_line_number = -1
        for i in range(loc_count):
            offset = i * 2
            line_number = self.read_int32()
            if line_number == 0:
                raise MarshalError("invalid source location")
            # OneIndexed::new used in Rust requires line != 0
            char_off_zero_indexed = self.read_int32()
            locations.append(
                SourceLocation(
                    line_number=line_number,
                    column_offset=char_off_zero_indexed + 1  # convert from zero-indexed
                )
            )
            if line_number != last_line_number and offset not in co_lnotab:
                co_lnotab[offset] = line_number
                last_line_number = line_number

        flags = self.read_int16()

        posonlyarg_count = self.read_int32()
        arg_count = self.read_int32()
        kwonlyarg_count = self.read_int32()

        # source_path
        src_len = self.t_int32(save_ref, bytes_for_s)
        source_path = self.read_string(src_len, False)

        first_line_raw = self.read_int32()
        first_line_number = None if first_line_raw == 0 else first_line_raw

        max_stackdepth = self.read_int32()

        obj_name_len = self.read_int32()
        obj_name = self.read_string(obj_name_len, False)

        if self.magic_int in (24881, 35310):
            qualname_len = self.read_int32()
            co_qualname = self.read_string(qualname_len, False)

        cell2arg_len = self.read_int32()
        if cell2arg_len != 0:
            cell2arg = []
            for _ in range(cell2arg_len):
                raw = self.read_int32()
                # convert raw (u32) to signed i32
                signed = raw if raw < (1 << 31) else raw - (1 << 32)
                cell2arg.append(signed)

        # constants
        const_count = self.read_int32()
        constants = []

        for _ in range(const_count):
            # deserialize_value must exist in your runtime; it is the
            # counterpart of the Rust deserialize_value()
            constants.append(self.r_object(bytes_for_s=bytes_for_s))

        def read_names():
            n = self.read_int32()
            out = []
            for _ in range(n):
                length = self.read_int32()
                out.append(self.read_string(length, False))
            return tuple(out)

        co_names = read_names()
        co_varnames = read_names()
        co_cellvars = read_names()
        co_freevars = read_names()
        co_nlocals = 0

        if self.magic_int in (24881, 35310):
            linetable_len = self.read_int32()
            co_linetable = self.read_slice(linetable_len)

            exceptiontable_len = self.read_int32()
            co_exceptiontable = self.read_slice(exceptiontable_len)
            return Code313Rust(
                co_argcount=arg_count,
                co_posonlyargcount=posonlyarg_count,
                co_kwonlyargcount=kwonlyarg_count,
                co_nlocals=co_nlocals,
                co_stacksize=max_stackdepth,
                co_flags=flags,
                co_code=co_code,
                co_consts=tuple(constants),
                co_names=co_names,
                co_varnames=co_varnames,
                co_filename=source_path,
                co_name=obj_name,
                co_qualname=co_qualname,
                co_firstlineno=first_line_number,
                co_linetable=co_linetable,
                co_freevars=co_freevars,
                co_cellvars=co_cellvars,
                co_exceptiontable=co_exceptiontable,
                version_triple=self.version_triple,
                locations=tuple(locations),
            )

        else:
            return Code312Rust(
                co_argcount=arg_count,
                co_posonlyargcount=posonlyarg_count,
                co_kwonlyargcount=kwonlyarg_count,
                co_lnotab=co_lnotab,
                co_stacksize=max_stackdepth,
                co_flags=flags,
                co_code=co_code,
                co_consts=tuple(constants),
                co_names=co_names,
                co_varnames=co_varnames,
                co_filename=source_path,
                co_name=obj_name,
                co_firstlineno=first_line_number,
                co_freevars=co_freevars,
                co_cellvars=co_cellvars,
                version_triple=self.version_triple,
                locations=tuple(locations),
            )



    def t_bigint(self, save_ref: bool=False, bytes_for_s: bool=False):
        len = self.read_int32()
        is_positive = len >= 0
        byte_data = self.read_slice(abs(len))
        value = int.from_bytes(byte_data, byteorder='little')
        return value if is_positive else -value

    def read_string(self, n: int, bytes_for_s: bool=False) -> Union[bytes, str]:
        s = self.read_slice(n)
        if not bytes_for_s:
            s = compat_str(s)
        return s



# (C) Copyright 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
CPython 3.12 bytecode opcodes

This is like Python 3.12's opcode.py with some classification
of stack usage and information for formatting instructions.
This has similar imformation as the opcodes in Python's opcode.py library.
"""

import xdis.cross_dis
import xdis.opcodes.opcode_3x.opcode_311 as opcode_311
from xdis.opcodes.base import (
    binary_op,
    call_op,
    cpython_implementation,
    def_op,
    finalize_opcodes,
    init_opdata,
    jrel_op,
    local_op,
    name_op,
    rm_op,
    update_pj3,
)
from xdis.opcodes.opcode_3x.opcode_311 import opcode_arg_fmt311, opcode_extended_fmt311

findlinestarts = xdis.cross_dis.findlinestarts
version_tuple = (3, 12)
python_implementation = cpython_implementation

loc = locals()

init_opdata(loc, opcode_311, version_tuple)


# extend opcodes to cover pseudo ops
loc["opname"].extend([f"<{i}>" for i in range(256, 267)])
loc["oppop"].extend([0] * 11)
loc["oppush"].extend([0] * 11)

# fmt: off
## These are removed / replaced since 3.11...
#           OP NAME                      OPCODE
#----------------------------------------------
rm_op(loc, "UNARY_POSITIVE"               , 10)
rm_op(loc, "PRINT_EXPR"                   , 70)

rm_op(loc, "LIST_TO_TUPLE"                , 82)
rm_op(loc, "IMPORT_STAR"                  , 84)
rm_op(loc, "ASYNC_GEN_WRAP"               , 87)
rm_op(loc, "PREP_RERAISE_STAR"            , 88)

rm_op(loc, "JUMP_IF_FALSE_OR_POP"         , 111)
rm_op(loc, "JUMP_IF_TRUE_OR_POP"          , 112)
rm_op(loc, "POP_JUMP_FORWARD_IF_FALSE"    , 114)
rm_op(loc, "POP_JUMP_FORWARD_IF_TRUE"     , 115)

rm_op(loc, "POP_JUMP_FORWARD_IF_NONE"     , 129)
rm_op(loc, "POP_JUMP_FORWARD_IF_NOT_NONE" , 128)
rm_op(loc, "LOAD_CLASSDEREF"              , 148)

rm_op(loc, "PRECALL"                      , 166)
rm_op(loc, "POP_JUMP_BACKWARD_IF_NOT_NONE", 173)
rm_op(loc, "POP_JUMP_BACKWARD_IF_NONE"    , 174)
rm_op(loc, "POP_JUMP_BACKWARD_IF_FALSE"   , 175)
rm_op(loc, "POP_JUMP_BACKWARD_IF_TRUE"    , 176)

rm_op(loc, "YIELD_VALUE"                  , 86)
rm_op(loc, "LOAD_METHOD"                  , 160)

## These are new since 3.11...
#            OP NAME                              OPCODE POP PUSH
#---------------------------------------------------------------
def_op(loc    , "INTERPRETER_EXIT"                 , 3  ,   1, 0)
def_op(loc    , "END_FOR"                          , 4  ,   2, 0)
def_op(loc    , "END_SEND"                         , 5  ,   1, 0)
def_op(loc    , "RESERVED"                         , 17 ,   0, 0)

binary_op(loc , "BINARY_SLICE"                     , 26 ,   2, 0)
binary_op(loc , "STORE_SLICE"                      , 27 ,   4, 0)

def_op(loc    , "CLEANUP_THROW"                    , 55 ,   2, 1)
def_op(loc    , "LOAD_LOCALS"                      , 87 ,   0, 1) # Pushes a reference to the locals of the current scope.
                                                                  # This is not a name op.
def_op(loc    , "RETURN_CONST"                     , 121,   0, 0)
local_op(loc  , "LOAD_FAST_CHECK"                  , 127,   0, 1)

jrel_op(loc   , "POP_JUMP_IF_FALSE"                , 114,   1, 0)
jrel_op(loc   , "POP_JUMP_IF_TRUE"                 , 115,   1, 0)
jrel_op(loc   , "POP_JUMP_IF_NOT_NONE"             , 128,   1, 0)
jrel_op(loc   , "POP_JUMP_IF_NONE"                 , 129,   1, 0)

def_op(loc    , "LOAD_SUPER_ATTR"                  , 141,   3, 1)
def_op(loc    , "LOAD_FAST_AND_CLEAR"              , 143,   0, 1)
def_op(loc    , "YIELD_VALUE"                      , 150,   1, 1)
def_op(loc    , "CALL_INTRINSIC_1"                 , 173,   1, 1)
def_op(loc    , "CALL_INTRINSIC_2"                 , 174,   2, 1)
def_op(loc    , "LOAD_FROM_DICT_OR_GLOBALS"        , 175,   1, 1)
def_op(loc    , "LOAD_FROM_DICT_OR_DEREF"          , 176,   1, 1)

#            OP NAME                               OPCODE POP PUSH
def_op(loc    , "INSTRUMENTED_LOAD_SUPER_ATTR"     , 237,   2, 0)
def_op(loc    , "INSTRUMENTED_POP_JUMP_IF_NONE"    , 238,   1, 1)
def_op(loc    , "INSTRUMENTED_POP_JUMP_IF_NOT_NONE", 239,   1, 1)
def_op(loc    , "INSTRUMENTED_RESUME"              , 240,   1, 1)
def_op(loc    , "INSTRUMENTED_CALL"                , 241,   1, 1)
def_op(loc    , "INSTRUMENTED_RETURN_VALUE"        , 242,   1, 0)
def_op(loc    , "INSTRUMENTED_YIELD_VALUE"         , 243,   1, 1)
call_op(loc   , "INSTRUMENTED_CALL_FUNCTION_EX"    , 244,   1, 1)
def_op(loc    , "INSTRUMENTED_JUMP_FORWARD"        , 245,   1, 1)
def_op(loc    , "INSTRUMENTED_JUMP_BACKWARD"       , 246,   1, 1)
def_op(loc    , "INSTRUMENTED_RETURN_CONST"        , 247,   1, 1)
def_op(loc    , "INSTRUMENTED_FOR_ITER"            , 248,   1, 1)
def_op(loc    , "INSTRUMENTED_POP_JUMP_IF_FALSE"   , 249,   1, 1)
def_op(loc    , "INSTRUMENTED_POP_JUMP_IF_TRUE"    , 250,   1, 1)
def_op(loc    , "INSTRUMENTED_END_FOR"             , 251,   2, 0)
def_op(loc    , "INSTRUMENTED_END_SEND"            , 252,   1, 0)
def_op(loc    , "INSTRUMENTED_INSTRUCTION"         , 253,   1, 1)
def_op(loc    , "INSTRUMENTED_LINE"                , 254,   1, 1)

#            OP NAME                              OPCODE POP PUSH
def_op(loc    , "SETUP_FINALLY"                    , 256,   0, 1)
def_op(loc    , "SETUP_CLEANUP"                    , 257,   0, 2)
def_op(loc    , "SETUP_WITH"                       , 258,   0, 1)
def_op(loc    , "POP_BLOCK"                        , 259,   0, 0)

jrel_op(loc   , "JUMP"                             , 260,   0, 0)
jrel_op(loc   , "JUMP_NO_INTERRUPT"                , 261,   0, 0)

name_op(loc   , "LOAD_METHOD"                      , 262,   0, 1)
def_op(loc    , "LOAD_SUPER_METHOD"                , 263,   1, 0)
def_op(loc    , "LOAD_ZERO_SUPER_METHOD"           , 264,   1, 0)
def_op(loc    , "LOAD_ZERO_SUPER_ATTR"             , 265,   1, 0)
def_op(loc    , "STORE_FAST_MAYBE_NULL"            , 266,   1, 0)

### update opinfo tables
loc["hasnargs"] = []
loc["hasstore"] = []
loc["hasvargs"] = []

loc["hasconst"].append(121)
loc["hasfree"].extend([148, 176])
# hasjrel removed 111 112 173 174 175 176 and added 260 261
loc["hasjrel"] = [93, 110, 114, 115, 123, 128, 129, 134, 140, 260, 261]
loc["haslocal"].extend([127, 143, 266])
# hasname removed 160 and added 141 175 262 234 264 265
loc["hasname"] = [90, 91, 95, 96, 97, 98, 101, 106, 108, 109, 116, 141, 175, 262, 263, 264, 265]
# new hasarg table
loc.update({"hasarg": [90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110,\
                       114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132,
                       133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 149, 150, 151, 152,
                       155, 156, 157, 162, 163, 164, 165, 171, 172, 173, 174, 175, 176, 237, 238, 239, 240, 241, 242,
                       243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 260, 261, 262, 263, 264, 265, 266]})

# new hasexc table
loc.update({"hasexc":[264, 265, 266]})

### add new arg formats
_intrinsic_1_descs = [
    "INTRINSIC_1_INVALID",
    "INTRINSIC_PRINT",
    "INTRINSIC_IMPORT_STAR",
    "INTRINSIC_STOPITERATION_ERROR",
    "INTRINSIC_ASYNC_GEN_WRAP",
    "INTRINSIC_UNARY_POSITIVE",
    "INTRINSIC_LIST_TO_TUPLE",
    "INTRINSIC_TYPEVAR",
    "INTRINSIC_PARAMSPEC",
    "INTRINSIC_TYPEVARTUPLE",
    "INTRINSIC_SUBSCRIPT_GENERIC",
    "INTRINSIC_TYPEALIAS",
]

_intrinsic_2_descs = [
    "INTRINSIC_2_INVALID",
    "INTRINSIC_PREP_RERAISE_STAR",
    "INTRINSIC_TYPEVAR_WITH_BOUND",
    "INTRINSIC_TYPEVAR_WITH_CONSTRAINTS",
    "INTRINSIC_SET_FUNCTION_TYPE_PARAMS",
]

def format_CALL_INTRINSIC_1(arg) -> str:
    return _intrinsic_1_descs[arg]

def format_CALL_INTRINSIC_2(arg) -> str:
    return _intrinsic_2_descs[arg]


opcode_extended_fmt = opcode_extended_fmt312 = opcode_extended_fmt311.copy()
opcode_arg_fmt = opcode_arg_fmt12 = opcode_arg_fmt311.copy()

### update arg formatting
opcode_arg_fmt312 = {
    **opcode_arg_fmt311,
    **{
        "CALL_INTRINSIC_1": format_CALL_INTRINSIC_1,
        "CALL_INTRINSIC_2": format_CALL_INTRINSIC_2,
    },
}

update_pj3(globals(), loc)
finalize_opcodes(loc)



# Copyright (c) 2015-2021, 2024-2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Graal version-independent Python object deserialization (unmarshal).

This is needed when the bytecode extracted is from
a different version than the currently-running Python.

When the running interpreter and the read-in bytecode are the same,
you can use Python's built-in ``marshal.loads()`` to produce a code
object.
"""

import marshal
from struct import unpack, unpack_from
from typing import Any, Dict, Tuple, Union

from xdis.codetype import to_portable
from xdis.magics import GRAAL3_MAGICS, magic_int2tuple
from xdis.unmarshal import (
    TYPE_ARRAY,
    TYPE_ASCII,
    TYPE_ASCII_INTERNED,
    TYPE_BINARY_COMPLEX,
    TYPE_BINARY_FLOAT,
    TYPE_COMPLEX,
    TYPE_DICT,
    TYPE_ELLIPSIS,
    TYPE_FALSE,
    TYPE_FLOAT,
    TYPE_FROZENSET,
    TYPE_GRAALPYTHON_CODE_UNIT,
    TYPE_INT,
    TYPE_INT64,
    TYPE_INTERNED,
    TYPE_LIST,
    TYPE_LONG,
    TYPE_NONE,
    TYPE_NULL,
    TYPE_REF,
    TYPE_SET,
    TYPE_SHORT_ASCII,
    TYPE_SHORT_ASCII_INTERNED,
    TYPE_SLICE,
    TYPE_SMALL_TUPLE,
    TYPE_STOPITER,
    TYPE_STRING,
    TYPE_STRINGREF,
    TYPE_TRUE,
    TYPE_TUPLE,
    TYPE_UNICODE,
    TYPE_UNKNOWN,
    VersionIndependentUnmarshaller,
)
from xdis.version_info import IS_GRAAL, PYTHON_VERSION_TRIPLE, version_tuple_to_str

# Graal Array types
TYPE_CODE_GRAAL = "C"
ARRAY_TYPE_BOOLEAN = "B"
ARRAY_TYPE_BYTE = "b"
ARRAY_TYPE_DOUBLE = "d"
ARRAY_TYPE_INT = "i"
ARRAY_TYPE_LONG = "l"
ARRAY_TYPE_OBJECT = "o"
ARRAY_TYPE_SHORT = "s"
ARRAY_TYPE_SLICE = ":"
ARRAY_TYPE_STRING = "S"


# The keys in the following dictionary are unmarshal codes, like "s",
# "c", "<", etc. The values of the dictionary are names of routines
# to call that do the data unmarshaling.
#
# Note: we could eliminate the parameters if this were all inside a
# class.  This might be good from an efficiency standpoint, and bad
# from a functional-programming standpoint. Pick your poison.
# EDIT: I'm choosing efficiency over functional programming.
UNMARSHAL_DISPATCH_TABLE = {
    TYPE_ARRAY: "graal_readArray",
    TYPE_ASCII: "ASCII",
    TYPE_ASCII_INTERNED: "ASCII_interned",
    TYPE_BINARY_COMPLEX: "binary_complex",
    TYPE_BINARY_FLOAT: "binary_float",
    TYPE_CODE_GRAAL: "code_graal",
    TYPE_COMPLEX: "complex",
    TYPE_DICT: "dict",
    TYPE_ELLIPSIS: "Ellipsis",
    TYPE_FALSE: "False",
    TYPE_FLOAT: "float",
    TYPE_FROZENSET: "frozenset",
    TYPE_GRAALPYTHON_CODE_UNIT: "graal_CodeUnit",
    TYPE_INT64: "int64",
    TYPE_INT: "int32",
    TYPE_INTERNED: "interned",
    TYPE_LIST: "list",
    TYPE_LONG: "long",
    TYPE_NONE: "None",
    TYPE_NULL: "C_NULL",
    TYPE_REF: "object_reference",
    TYPE_SET: "set",
    TYPE_SHORT_ASCII: "short_ASCII",
    TYPE_SHORT_ASCII_INTERNED: "short_ASCII_interned",
    TYPE_SLICE: "slice",
    TYPE_SMALL_TUPLE: "small_tuple",
    TYPE_STOPITER: "stopIteration",
    TYPE_STRING: "string",
    TYPE_STRINGREF: "python2_string_reference",
    TYPE_TRUE: "True",
    TYPE_TUPLE: "tuple",
    TYPE_UNICODE: "unicode",
    TYPE_UNKNOWN: "unknown",
}

# Used by graal
JAVA_MARSHAL_SHIFT = 15
JAVA_MARSHAL_BASE = 1 << JAVA_MARSHAL_SHIFT

class VersionIndependentUnmarshallerGraal(VersionIndependentUnmarshaller):
    def __init__(self, fp, magic_int, bytes_for_s, code_objects={}) -> None:
        """
        Marshal versions:
            2: [2.5, 3.4a0) (self.magic_int: 62071 until 3250)
            3: [3.4a0, 3.4a3) (self.magic_int: 3250 until 3280)
            4: [3.4a3, 3.13) (self.magic_int: 3280 onwards)
            5: [3.14, current) (self.magic_int: circa 3608)

        In Python 3, a ``bytes`` type is used for strings.
        """
        self.fp = fp
        self.magic_int = magic_int
        self.code_objects = code_objects

        # Save a list of offsets in the bytecode file where code
        # objects starts.
        self.code_to_file_offsets = {}

        # It is helpful to save the order in sets, frozensets and dictionary keys,
        # so that on writing a bytecode file we can duplicate this order.
        self.collection_order: Dict[Union[set, frozenset, dict], Tuple[Any]] = {}

        self.bytes_for_s = bytes_for_s
        self.version_triple = magic_int2tuple(self.magic_int)
        if self.version_triple >= (3, 14):
            self.marshal_version = 5
        elif (3, 14) > self.version_triple >= (3, 4):
            if self.magic_int in (3250, 3260, 3270):
                self.marshal_version = 3
            else:
                self.marshal_version = 4
        elif (3, 4) > self.version_triple >= (2, 5):
            self.marshal_version = 2
        else:
            assert False, f"version {version_tuple_to_str(self.version.triple)} is not a graal version"

        self.intern_strings = []
        self.intern_objects = []
        self.is_graal = True
        self.is_pypy = False
        self.is_rust = False

        # Graal code data seems to be split in two places.
        # The outer part is a TYPE_GRAALPYTHON_CODE,
        # while the inner part is in TYPE_GRAALPYTHON_CODE_UNIT
        # Save the out information so that the inner information
        # can use this when building a code type.

        # in a TYPE_GRAAL
        self.graal_code_info = {}

        self.UNMARSHAL_DISPATCH_TABLE = UNMARSHAL_DISPATCH_TABLE


    # Python equivalents for graal unmarshal routines.
    def t_graal_readArray(self, save_ref: bool, bytes_for_s: bool) -> tuple:
        """
        Python equivalent of Python Graal's readArray() from
        MarshalModuleBuiltins.java

        Array object unmarshalling read routine.  Reads from self.fp
        the next byte which is a key in ARRAY_TYPE defined above.

        Parameter save_ref indicates whether to save the resulting object in
        our internal object cache.
        """
        byte1 = ord(self.fp.read(1))

        marshal_type = chr(byte1)

        # print(marshal_type)  # debug

        match marshal_type:
            case "B":
                ret = self.graal_readBooleanArray()
            # case "b":
            #     return "OK"
            case "d":
                ret = self.graal_readDoubleArray()
            case "i":
                ret = self.graal_readIntArray()
            case "o":
                ret = self.graal_readObjectArray()
            case "l":
                ret = self.graal_readLongArray()
            # case "s":
            #     return "Internal server error"
            case "S":
                ret = self.graal_readStringArray()
            case _:
                # The underscore '_' acts as a wildcard
                # It matches anything if no previous case did (the 'default' case)
                print(f"XXX Whoah {marshal_type}")
                ret = tuple()
        if save_ref:
            if isinstance(ret, tuple):
                self.intern_objects += list(ret)
            else:
                self.intern_objects.append(ret)
        return ret

    def graal_readBigInteger(self):
        """
        Reads a marshaled big integer from the input stream.
        """
        negative = False
        sz = self.read_uint32() # Get the size in shorts
        if sz < 0:
            negative = True
            sz = -sz

        # size is in shorts, convert to size in bytes
        sz *= 2

        data = bytes([self.graal_readByte() for _ in range(sz)])

        i = 0

        # Read the first 2 bytes as a 16-bit signed integer (short)
        # '>h' specifies big-endian signed short
        digit = unpack_from('>h', data, i)[0]
        i += 2

        # Python int handles arbitrarily large numbers
        result = digit

        while i < sz:
            # Calculate the power based on the number of shorts processed so far
            power = i // 2
            # Read the next 2 bytes as a 16-bit signed integer
            digit = unpack_from('>h', data, i)[0]
            i += 2

            # In Python, int supports all these operations directly
            # The Java code effectively reconstructs the number using base 2^16 (MARSHAL_BASE)
            term = digit * (JAVA_MARSHAL_BASE ** power)
            result += term

        if negative:
            return -result
        else:
            return result

    def graal_readBooleanArray(self) -> tuple[bool, ...]:
        """
        Python equivalent of Python Graal's readBooleanArray() from
        MarshalModuleBuiltins.java
        """
        length: int = self.read_uint32()
        return tuple([bool(self.graal_readByte()) for _ in range(length)])

    def graal_readByte(self) -> int:
        """
        Python equivalent of Python Graal's readBytes() from
        MarshalModuleBuiltins.java
        """
        return ord(unpack("c", self.fp.read(1))[0])

    def graal_readBytes(self) -> bytes:
        """
        Python equivalent of Python Graal's readBytes() from
        MarshalModuleBuiltins.java
        """
        length: int = self.read_uint32()
        return bytes([self.graal_readByte() for _ in range(length)])

    def graal_readDoubleArray(self) -> tuple[float, ...]:
        """
        Python equivalent of Python Graal's readDoubleArray() from
        MarshalModuleBuiltins.java
        """
        length: int = self.read_uint32()
        return tuple([self.read_float() for _ in range(length)])

    def graal_readIntArray(self) -> tuple[int, ...]:
        """
        Python equivalent of Python Graal's readIntArray() from
        MarshalModuleBuiltins.java
        """
        length: int = self.read_uint32()
        return tuple([self.read_int32() for _ in range(length)])

    def graal_readLongArray(self) -> tuple[int, ...]:
        """
        Python equivalent of Python Graal's readLong() from
        MarshalModuleBuiltins.java
        """
        length: int = int(unpack("<i", self.fp.read(4))[0])
        return tuple([self.read_float() for _ in range(length)])

    def graal_readObjectArray(self) -> tuple:
        """
        Python equivalent of Python Graal's readObjectArray() from
        MarshalModuleBuiltins.java
        """

        length: int = int(unpack("<i", self.fp.read(4))[0])
        # # Debug code
        # result = []
        # for i in range(length):
        #     try:
        #         result.append(self.r_object(bytes_for_s=False))
        #     except:
        #         breakpoint()
        # return tuple(result)
        return tuple([self.r_object(bytes_for_s=False) for _ in range(length)])

    def graal_readString(self) -> str:
        """
        Python equvalent of Python Graal's readString() from
        MarshalModuleBuiltins.java
        """
        strsize: int = self.read_uint32()
        return self.fp.read(strsize).decode("utf-8", errors="ignore")

    def graal_readStringArray(self) -> tuple[str, ...]:
        """
        Python equvalent of Python Graal's readObjectArray() from
        MarshalModuleBuiltins.java
        """
        length: int = self.read_uint32()
        return tuple([self.graal_readString() for _ in range(length)])

    def graal_readSparseTable(self) -> Dict[int, tuple]:
        """
        Python equvalent of Python Graal's readObjectArray() from
        MarshalModuleBuiltins.java
        """
        self.read_uint32()  # the length return value isn't used.
        table = {}  # new int[length][];
        while True:
            i = self.read_int32()
            if i == -1:
                return table
            table[i] = self.graal_readIntArray()

    def t_code_graal(self, save_ref, bytes_for_s: bool = False):
        """
        Graal Python code. This has fewer fields than Python
        code. In particular, instructions are JVM bytecode.
        """

        # Go back one byte to TYPE_CODE "C" or "C" with the FLAG_REF
        # set.
        code_offset_in_file = self.fp.tell() - 1

        # Below, the value None (slot for a code object value), will
        # be replaced by the actual code in variable `ret` after it
        # has been built.
        ret, i = self.r_ref_reserve(None, save_ref)

        self.version_triple = magic_int2tuple(self.magic_int)

        # This is graal 312 Java code for writing a bytecode file
        # after the file header info had been read:
        #   writeByte(TYPE_GRAALPYTHON_CODE | flag);
        #   writeString(c.getFilename());
        #   writeInt(c.getFlags());
        #   writeBytes(c.getCodestring());
        #   writeInt(c.getFirstLineNo());
        #   byte[] lnotab = c.getLinetable();
        #   if (lnotab == null) {
        #         lnotab = PythonUtils.EMPTY_BYTE_ARRAY;
        #   }
        #   writeBytes(lnotab);

        self.graal_code_info["co_filename"] = self.graal_readString()
        self.graal_code_info["co_flags"] = self.t_int32(False, bytes_for_s=False)
        co_codeunit_position = self.graal_code_info["co_codeunit_position"] = (
            self.fp.tell() + 4
        )
        co_codeunit_string = self.graal_readBytes()
        # Determine if we are Graal 3.10, 3.11, or 3.12 ...
        if chr(co_codeunit_string[0]) != TYPE_GRAALPYTHON_CODE_UNIT:
            # Determine if we are Graal 3.10, 3.11
            co_codeunit_string = b"U" + co_codeunit_string
            if self.version_triple != (3, 10, 8):
                self.version_triple = (3, 11, 7)
        else:
            self.version_triple = (3, 12, 8)

        self.graal_code_info["co_firstlineno"] = self.t_int32(False, bytes_for_s=False)
        self.graal_code_info["co_lnotab"] = self.t_string(False, bytes_for_s=True)

        # Go back to code position and parse CodeUnit
        saved_position = self.fp.tell()
        self.fp.seek(co_codeunit_position)
        if self.version_triple == (3, 12, 8):
            code = self.r_object(bytes_for_s=False)
        else:
            code = self.t_graal_CodeUnit(save_ref=False, bytes_for_s=False)
        assert self.graal_code_info["co_flags"] == code.co_flags

        # FIXME: add an assert self.fp.tell() has advanced to save_position?
        self.fp.seek(saved_position)

        code.graal_instr_str = ""
        if IS_GRAAL and PYTHON_VERSION_TRIPLE == self.version_triple:
            try:
                graal_instr_str = str(marshal.loads(co_codeunit_string))
            except Exception:
                pass
            else:
                code.graal_instr_str = graal_instr_str

        self.code_to_file_offsets[code] = (
            code_offset_in_file,
            self.code_to_file_offsets[code][0],
        )
        self.code_objects[str(code)] = code

        return self.r_ref_insert(code, i)

    def t_graal_CodeUnit(self, save_ref, bytes_for_s: bool = False):
        """
        Graal Python code. This has fewer fields than Python
        code. Graal Python Bytecode is similar to Python Bytecode for
        a given version, things are a little different, especially when it comes
        to collections. Graal has intructions for homogeneous arrays.

        Note that Graal's BytecodeCodeUnit has a dumpBytecode() method
        to show assembly instructions.
        """

        graal_bytecode_version = self.graal_readByte()
        assert (21000 + graal_bytecode_version * 10) in GRAAL3_MAGICS
        if graal_bytecode_version in (26,):
            self.version_triple = (3, 8, 5)

        other_fields: Dict = {}

        # This is Java code for how a CodeUnit (type "U") is read
        # TruffleString name = readString();
        # TruffleString qualname = readString();
        # int argCount = readInt();
        # int kwOnlyArgCount = readInt();
        # int positionalOnlyArgCount = readInt();
        # int stacksize = readInt();
        # byte[] code = readBytes();
        # byte[] srcOffsetTable = readBytes();
        # int flags = readInt();

        co_name = self.graal_readString()
        co_qualname = self.graal_readString()
        co_argcount = self.read_uint32()
        co_kwonlyargcount = self.read_uint32()
        co_posonlyargcount = self.read_uint32()

        co_stacksize = self.read_uint32()
        co_code_offset_in_file = self.fp.tell()
        co_code = self.graal_readBytes()
        other_fields["srcOffsetTable"] = self.graal_readBytes()
        co_flags = self.read_uint32()

        # writeStringArray(code.names);
        # writeStringArray(code.varnames);
        # writeStringArray(code.cellvars);
        # writeStringArray(code.freevars);

        co_names = self.graal_readStringArray()
        co_varnames = self.graal_readStringArray()
        co_cellvars = self.graal_readStringArray()
        co_freevars = self.graal_readStringArray()

        # int[] cell2arg = readIntArray();
        # if (cell2arg.length == 0) {
        #     cell2arg = null;
        #     }
        # Object[] constants = readObjectArray();

        other_fields["cell2arg"] = self.graal_readIntArray()
        co_consts = self.graal_readObjectArray()

        # The data from the below is not used, but we run the
        # the extraction to keep the self.fp location where it should for
        # the situation that we have code objects inside the codes' co_consts
        # table marshaled as an ObjectArray.

        # long[] primitiveConstants = readLongArray();
        # int[] exceptionHandlerRanges = readIntArray();
        # int conditionProfileCount = readInt();
        # int startLine = readInt();
        # int startColumn = readInt();
        # int endLine = readInt();
        # int endColumn = readInt();
        # byte[] outputCanQuicken = readBytes();
        # byte[] variableShouldUnbox = readBytes();
        # int[][] generalizeInputsMap = readSparseTable();
        # int[][] generalizeVarsMap = readSparseTable();

        other_fields["primitiveConstants"] = self.graal_readLongArray()
        other_fields["exception_handler_ranges"] = self.graal_readIntArray()
        other_fields["condition_profileCount"] = self.read_uint32()
        other_fields["startLine"] = self.read_uint32()
        other_fields["startColumn"] = self.read_uint32()
        other_fields["endLine"] = self.read_uint32()
        other_fields["endColumn"] = self.read_uint32()
        other_fields["outputCanQuicken"] = self.graal_readBytes()
        other_fields["variableShouldUnbox"] = self.graal_readBytes()
        other_fields["generalizeInputsMap"] = self.graal_readSparseTable()
        other_fields["generalizeVarsMap"] = self.graal_readSparseTable()

        code = to_portable(
            co_argcount=co_argcount,
            co_posonlyargcount=co_posonlyargcount,
            co_kwonlyargcount=co_kwonlyargcount,
            co_nlocals=0,
            co_stacksize=co_stacksize,
            co_flags=co_flags,
            co_code=co_code,
            co_consts=co_consts,
            co_names=co_names,
            co_varnames=co_varnames,
            co_filename=self.graal_code_info["co_filename"],
            co_name=co_name,
            co_qualname=co_qualname,
            co_firstlineno=self.graal_code_info["co_firstlineno"],
            co_lnotab=self.graal_code_info["co_lnotab"],
            co_freevars=co_freevars,
            co_cellvars=co_cellvars,
            co_exceptiontable="",
            version_triple=self.version_triple,
            collection_order=self.collection_order,
            reference_objects=set(),
            other_fields=other_fields,
        )

        self.code_to_file_offsets[code] = tuple(
            [self.graal_code_info["co_codeunit_position"], co_code_offset_in_file]
        )

        # if graal_bytecode_version == 26:
        #     from xdis.bytecode_graal import get_instructions_bytes_graal
        #     from xdis.opcodes import opcode_38graal
        #     get_instructions_bytes_graal(code, opcode_38graal)
        return code



# (C) Copyright 2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
CPython 3.13 bytecode opcodes.

This is like Python 3.13's opcode.py with some classification
of stack usage and information for formatting instructions.
This has similar imformation as the opcodes in Python's opcode.py library.
"""

from typing import Optional, Tuple

import xdis.opcodes.opcode_3x.opcode_312 as opcode_312
from xdis.opcodes.base import (
    VARYING_STACK_INT,
    call_op,
    cpython_implementation,
    def_op,
    finalize_opcodes,
    free_op,
    init_opdata,
    local_op,
    nargs_op,
    rm_op,
    update_pj3,
)
from xdis.opcodes.format.extended import NULL_EXTENDED_OP, get_arglist
from xdis.opcodes.opcode_pypy.opcode_36pypy import format_CALL_METHOD

version_tuple = (3, 13)
python_implementation = cpython_implementation

loc = locals()

init_opdata(loc, opcode_312, version_tuple)

# extend op tables for new pseudo ops
loc["opname"].extend(["<267>"])
loc["oppop"].extend([0])
loc["oppush"].extend([0])

# have argument changed in 3.13
HAVE_ARGUMENT = 44

# fmt: off
## Lots of ops changed opcodes in 3.13 so this is long...
## These are removed or replaced since 3.12...
#           OP NAME                      OPCODE
#----------------------------------------------
rm_op(loc, "POP_TOP"                          , 1)
rm_op(loc, "PUSH_NULL"                        , 2)
rm_op(loc, "INTERPRETER_EXIT"                 , 3)
rm_op(loc, "END_FOR"                          , 4)
rm_op(loc, "END_SEND"                         , 5)
rm_op(loc, "NOP"                              , 9)
rm_op(loc, "UNARY_NEGATIVE"                   , 11)
rm_op(loc, "UNARY_NOT"                        , 12)
rm_op(loc, "UNARY_INVERT"                     , 15)
rm_op(loc, "BINARY_SUBSCR"                    , 25)
rm_op(loc, "BINARY_SLICE"                     , 26)
rm_op(loc, "STORE_SLICE"                      , 27)
rm_op(loc, "GET_LEN"                          , 30)
rm_op(loc, "MATCH_MAPPING"                    , 31)
rm_op(loc, "MATCH_SEQUENCE"                   , 32)
rm_op(loc, "MATCH_KEYS"                       , 33)
rm_op(loc, "PUSH_EXC_INFO"                    , 35)
rm_op(loc, "CHECK_EXC_MATCH"                  , 36)
rm_op(loc, "CHECK_EG_MATCH"                   , 37)
rm_op(loc, "WITH_EXCEPT_START"                , 49)
rm_op(loc, "GET_AITER"                        , 50)
rm_op(loc, "GET_ANEXT"                        , 51)
rm_op(loc, "BEFORE_ASYNC_WITH"                , 52)
rm_op(loc, "BEFORE_WITH"                      , 53)
rm_op(loc, "END_ASYNC_FOR"                    , 54)
rm_op(loc, "CLEANUP_THROW"                    , 55)
rm_op(loc, "DELETE_SUBSCR"                    , 61)
rm_op(loc, "GET_ITER"                         , 68)
rm_op(loc, "GET_YIELD_FROM_ITER"              , 69)
rm_op(loc, "LOAD_BUILD_CLASS"                 , 71)
rm_op(loc, "LOAD_ASSERTION_ERROR"             , 74)
rm_op(loc, "RETURN_GENERATOR"                 , 75)
rm_op(loc, "RETURN_VALUE"                     , 83)
rm_op(loc, "SETUP_ANNOTATIONS"                , 85)
rm_op(loc, "LOAD_LOCALS"                      , 87)
rm_op(loc, "POP_EXCEPT"                       , 89)
rm_op(loc, "STORE_NAME"                       , 90)
rm_op(loc, "DELETE_NAME"                      , 91)
rm_op(loc, "UNPACK_SEQUENCE"                  , 92)
rm_op(loc, "FOR_ITER"                         , 93)
rm_op(loc, "UNPACK_EX"                        , 94)
rm_op(loc, "STORE_ATTR"                       , 95)
rm_op(loc, "DELETE_ATTR"                      , 96)
rm_op(loc, "STORE_GLOBAL"                     , 97)
rm_op(loc, "DELETE_GLOBAL"                    , 98)
rm_op(loc, "SWAP"                             , 99)
rm_op(loc, "LOAD_CONST"                       , 100)
rm_op(loc, "LOAD_NAME"                        , 101)
rm_op(loc, "BUILD_TUPLE"                      , 102)
rm_op(loc, "BUILD_LIST"                       , 103)
rm_op(loc, "BUILD_SET"                        , 104)
rm_op(loc, "BUILD_MAP"                        , 105)
rm_op(loc, "COMPARE_OP"                       , 107)
rm_op(loc, "IMPORT_NAME"                      , 108)
rm_op(loc, "IMPORT_FROM"                      , 109)
rm_op(loc, "JUMP_FORWARD"                     , 110)
rm_op(loc, "POP_JUMP_IF_FALSE"                , 114)
rm_op(loc, "POP_JUMP_IF_TRUE"                 , 115)
rm_op(loc, "LOAD_GLOBAL"                      , 116)
rm_op(loc, "IS_OP"                            , 117)
rm_op(loc, "CONTAINS_OP"                      , 118)
rm_op(loc, "RERAISE"                          , 119)
rm_op(loc, "COPY"                             , 120)
rm_op(loc, "RETURN_CONST"                     , 121)
rm_op(loc, "BINARY_OP"                        , 122)
rm_op(loc, "SEND"                             , 123)
rm_op(loc, "LOAD_FAST"                        , 124)
rm_op(loc, "STORE_FAST"                       , 125)
rm_op(loc, "DELETE_FAST"                      , 126)
rm_op(loc, "LOAD_FAST_CHECK"                  , 127)
rm_op(loc, "POP_JUMP_IF_NOT_NONE"             , 128)
rm_op(loc, "POP_JUMP_IF_NONE"                 , 129)
rm_op(loc, "RAISE_VARARGS"                    , 130)
rm_op(loc, "GET_AWAITABLE"                    , 131)
rm_op(loc, "MAKE_FUNCTION"                    , 132)
rm_op(loc, "BUILD_SLICE"                      , 133)
rm_op(loc, "JUMP_BACKWARD_NO_INTERRUPT"       , 134)
rm_op(loc, "MAKE_CELL"                        , 135)
rm_op(loc, "LOAD_DEREF"                       , 137)
rm_op(loc, "STORE_DEREF"                      , 138)
rm_op(loc, "DELETE_DEREF"                     , 139)
rm_op(loc, "JUMP_BACKWARD"                    , 140)
rm_op(loc, "LOAD_SUPER_ATTR"                  , 141)
rm_op(loc, "CALL_FUNCTION_EX"                 , 142)
rm_op(loc, "LOAD_FAST_AND_CLEAR"              , 143)
rm_op(loc, "EXTENDED_ARG"                     , 144)
rm_op(loc, "LIST_APPEND"                      , 145)
rm_op(loc, "SET_ADD"                          , 146)
rm_op(loc, "MAP_ADD"                          , 147)
rm_op(loc, "YIELD_VALUE"                      , 150)
rm_op(loc, "RESUME"                           , 151)
rm_op(loc, "MATCH_CLASS"                      , 152)
rm_op(loc, "FORMAT_VALUE"                     , 155)
rm_op(loc, "BUILD_CONST_KEY_MAP"              , 156)
rm_op(loc, "BUILD_STRING"                     , 157)
rm_op(loc, "LIST_EXTEND"                      , 162)
rm_op(loc, "SET_UPDATE"                       , 163)
rm_op(loc, "DICT_MERGE"                       , 164)
rm_op(loc, "DICT_UPDATE"                      , 165)
rm_op(loc, "CALL"                             , 171)
rm_op(loc, "KW_NAMES"                         , 172)
rm_op(loc, "CALL_INTRINSIC_1"                 , 173)
rm_op(loc, "CALL_INTRINSIC_2"                 , 174)
rm_op(loc, "LOAD_FROM_DICT_OR_GLOBALS"        , 175)
rm_op(loc, "LOAD_FROM_DICT_OR_DEREF"          , 176)
rm_op(loc, "LOAD_CLOSURE"                     , 136)  # now a psuedo-instruction replaced with LOAD_FAST
rm_op(loc, "COPY_FREE_VARS"                   , 149)
rm_op(loc, "INSTRUMENTED_LOAD_SUPER_ATTR"     , 237)
rm_op(loc, "INSTRUMENTED_POP_JUMP_IF_NONE"    , 238)
rm_op(loc, "INSTRUMENTED_POP_JUMP_IF_NOT_NONE", 239)
rm_op(loc, "INSTRUMENTED_RESUME"              , 240)
rm_op(loc, "INSTRUMENTED_CALL"                , 241)
rm_op(loc, "INSTRUMENTED_RETURN_VALUE"        , 242)
rm_op(loc, "INSTRUMENTED_YIELD_VALUE"         , 243)
rm_op(loc, "INSTRUMENTED_CALL_FUNCTION_EX"    , 244)
rm_op(loc, "INSTRUMENTED_JUMP_FORWARD"        , 245)
rm_op(loc, "INSTRUMENTED_JUMP_BACKWARD"       , 246)
rm_op(loc, "INSTRUMENTED_RETURN_CONST"        , 247)
rm_op(loc, "INSTRUMENTED_FOR_ITER"            , 248)
rm_op(loc, "INSTRUMENTED_POP_JUMP_IF_FALSE"   , 249)
rm_op(loc, "INSTRUMENTED_END_FOR"             , 251)
rm_op(loc, "INSTRUMENTED_END_SEND"            , 252)
rm_op(loc, "INSTRUMENTED_INSTRUCTION"         , 253)
rm_op(loc, "SETUP_FINALLY"                    , 256)
rm_op(loc, "SETUP_CLEANUP"                    , 257)
rm_op(loc, "POP_BLOCK"                        , 259)
rm_op(loc, "JUMP"                             , 260)
rm_op(loc, "JUMP_NO_INTERRUPT"                , 261)
rm_op(loc, "LOAD_METHOD"                      , 262)
rm_op(loc, "LOAD_SUPER_METHOD"                , 263)
rm_op(loc, "LOAD_ZERO_SUPER_METHOD"           , 264)
rm_op(loc, "LOAD_ZERO_SUPER_ATTR"             , 265)
rm_op(loc, "STORE_FAST_MAYBE_NULL"            , 266)

## new or redefined ops
#            OP NAME                              OPCODE POP PUSH
# ---------------------------------------------------------------
def_op(loc, "BEFORE_ASYNC_WITH"                , 1   , 0 , 1)
def_op(loc, "BEFORE_WITH"                      , 2   , 0 , 1)
def_op(loc, "BINARY_SLICE"                     , 4   , 2 , 0)
def_op(loc, "BINARY_SUBSCR"                    , 5   , 2 , 1)
def_op(loc, "CHECK_EG_MATCH"                   , 6   , 0 , 0)
def_op(loc, "CHECK_EXC_MATCH"                  , 7   , 0 , 0)
def_op(loc, "CLEANUP_THROW"                    , 8   , 2 , 1)
def_op(loc, "DELETE_SUBSCR"                    , 9   , 2 , 0)
def_op(loc, "END_ASYNC_FOR"                    , 10  , 2 , 0)
def_op(loc, "END_FOR"                          , 11  , 1 , 0)
def_op(loc, "END_SEND"                         , 12  , 1 , 0)
def_op(loc, "EXIT_INIT_CHECK"                  , 13  , 1 , 0)
def_op(loc, "FORMAT_WITH_SPEC"                 , 15  , 1 , 0)
def_op(loc, "GET_AITER"                        , 16  , 1 , 1)
def_op(loc, "GET_ANEXT"                        , 18  , 0 , 1)
def_op(loc, "GET_ITER"                         , 19  , 1 , 1)
def_op(loc, "GET_LEN"                          , 20  , 0 , 1)
def_op(loc, "GET_YIELD_FROM_ITER"              , 21  , 1 , 1)
def_op(loc, "INTERPRETER_EXIT"                 , 22  , 1 , 0)
def_op(loc, "LOAD_ASSERTION_ERROR"             , 23  , 0 , 1)
def_op(loc, "LOAD_BUILD_CLASS"                 , 24  , 0 , 1)
def_op(loc, "LOAD_LOCALS"                      , 25  , 0 , 1)  # Pushes a reference to the locals of the current scope.
                                                               # This is not a name op.
nargs_op(loc, "MAKE_FUNCTION"                  , 26  , VARYING_STACK_INT, 1)
def_op(loc, "MATCH_KEYS"                       , 27  , 0 , 1)
def_op(loc, "MATCH_MAPPING"                    , 28  , 0 , 1)
def_op(loc, "MATCH_SEQUENCE"                   , 29  , 0 , 1)
def_op(loc, "NOP"                              , 30  , 0 , 0)
def_op(loc, "POP_EXCEPT"                       , 31  , 1 , 0)
def_op(loc, "POP_TOP"                          , 32  , 1 , 0)
def_op(loc, "PUSH_EXC_INFO"                    , 33  , 0 , 1)
def_op(loc, "PUSH_NULL"                        , 34  , 0 , 1)
def_op(loc, "RETURN_GENERATOR"                 , 35  , 0 , 1)
def_op(loc, "RETURN_VALUE"                     , 36  , 1 , 0)
def_op(loc, "SETUP_ANNOTATIONS"                , 37  , 1 , 1)
def_op(loc, "STORE_SLICE"                      , 38  , 4 , 0)
def_op(loc, "STORE_SUBSCR"                     , 39  , 3 , 0)
def_op(loc, "UNARY_INVERT"                     , 41  , 1 , 1)
def_op(loc, "UNARY_NEGATIVE"                   , 42  , 1 , 1)
def_op(loc, "UNARY_NOT"                        , 43  , 1 , 1)
def_op(loc, "WITH_EXCEPT_START"                , 44  , 0 , 1)
def_op(loc, "BINARY_OP"                        , 45  , 2 , 1)
def_op(loc, "BUILD_CONST_KEY_MAP"              , 46  , -2, 1)
def_op(loc, "BUILD_LIST"                       , 47  , -1, 1)
def_op(loc, "BUILD_MAP"                        , 48  , -1, -1)
def_op(loc, "BUILD_SET"                        , 49  , -1, 1)
def_op(loc, "BUILD_SLICE"                      , 50  , 2 , 1)
def_op(loc, "BUILD_STRING"                     , 51  , -2, 2)
def_op(loc, "BUILD_TUPLE"                      , 52  , -1, 1)
call_op(loc, "CALL"                            , 53  , 1 , 0)
call_op(loc, "CALL_FUNCTION_EX"                , 54  , -2, 1)
call_op(loc, "CALL_INTRINSIC_1"                , 55  , 1 , 1)
call_op(loc, "CALL_INTRINSIC_2"                , 56  , 2 , 1)
def_op(loc, "COMPARE_OP"                       , 58  , 2 , 1)
def_op(loc, "CONTAINS_OP"                      , 59  , 2 , 1)
def_op(loc, "COPY"                             , 61  , 0 , 1)
def_op(loc, "COPY_FREE_VARS"                   , 62  , 0 , 0)
def_op(loc, "DELETE_ATTR"                      , 63  , 1 , 0)
def_op(loc, "DELETE_DEREF"                     , 64  , 0 , 0)
def_op(loc, "DELETE_FAST"                      , 65  , 0 , 0)
def_op(loc, "DELETE_GLOBAL"                    , 66  , 0 , 0)
def_op(loc, "DELETE_NAME"                      , 67  , 0 , 0)
def_op(loc, "DICT_MERGE"                       , 68  , 2 , 1)
def_op(loc, "DICT_UPDATE"                      , 69  , 2 , 1)
def_op(loc, "ENTER_EXECUTOR"                   , 70  , 0 , 0)
def_op(loc, "EXTENDED_ARG"                     , 71  , 0 , 0)
def_op(loc, "FOR_ITER"                         , 72  , 0 , 1)
def_op(loc, "GET_AWAITABLE"                    , 73  , 0 , 0)
def_op(loc, "IMPORT_FROM"                      , 74  , 0 , 1)
def_op(loc, "IMPORT_NAME"                      , 75  , 2 , 1)
def_op(loc, "IS_OP"                            , 76  , 2 , 1)
def_op(loc, "JUMP_BACKWARD"                    , 77  , 0 , 0)
def_op(loc, "JUMP_BACKWARD_NO_INTERRUPT"       , 78  , 0 , 0)
def_op(loc, "JUMP_FORWARD"                     , 79  , 0 , 0)
def_op(loc, "LIST_APPEND"                      , 80  , 2 , 1)
def_op(loc, "LIST_EXTEND"                      , 81  , 2 , 1)
def_op(loc, "LOAD_ATTR"                        , 82  , 1 , 1)
def_op(loc, "LOAD_CONST"                       , 83  , 0 , 1)
free_op(loc, "LOAD_DEREF"                      , 84  , 0 , 1)
def_op(loc, "LOAD_FAST"                        , 85  , 0 , 1)
def_op(loc, "LOAD_FAST_AND_CLEAR"              , 86  , 0 , 1)
local_op(loc, "LOAD_FAST_CHECK"                , 87  , 0 , 1)
def_op(loc, "LOAD_FAST_LOAD_FAST"              , 88  , 0 , 2)
def_op(loc, "LOAD_FROM_DICT_OR_DEREF"          , 89  , 1 , 1)
def_op(loc, "LOAD_FROM_DICT_OR_GLOBALS"        , 90  , 1 , 1)
def_op(loc, "LOAD_GLOBAL"                      , 91  , 0 , 1)
def_op(loc, "LOAD_NAME"                        , 92  , 0 , 1)
def_op(loc, "LOAD_SUPER_ATTR"                  , 93  , 3 , 1)
free_op(loc, "MAKE_CELL"                       , 94  , 0 , 0)
def_op(loc, "MAP_ADD"                          , 95  , 3 , 1)
def_op(loc, "MATCH_CLASS"                      , 96  , 3 , 1)
def_op(loc, "POP_JUMP_IF_FALSE"                , 97  , 1 , 0)
def_op(loc, "POP_JUMP_IF_NONE"                 , 98  , 1 , 0)
def_op(loc, "POP_JUMP_IF_NOT_NONE"             , 99  , 1 , 0)
def_op(loc, "POP_JUMP_IF_TRUE"                 , 100 , 1 , 0)
def_op(loc, "RAISE_VARARGS"                    , 101 , -1, 1)
def_op(loc, "RERAISE"                          , 102 , 1 , 0)
def_op(loc, "RETURN_CONST"                     , 103 , 0 , 0)
def_op(loc, "SEND"                             , 104 , 0 , 0)
def_op(loc, "SET_ADD"                          , 105 , 1 , 0)
def_op(loc, "SET_UPDATE"                       , 107 , 2 , 1)
def_op(loc, "STORE_ATTR"                       , 108 , 2 , 0)
def_op(loc, "STORE_DEREF"                      , 109 , 1 , 0)
def_op(loc, "STORE_FAST"                       , 110 , 1 , 0)
def_op(loc, "STORE_FAST_LOAD_FAST"             , 111 , 1 , 1)
def_op(loc, "STORE_FAST_STORE_FAST"            , 112 , 2 , 0)
def_op(loc, "STORE_GLOBAL"                     , 113 , 1 , 0)
def_op(loc, "STORE_NAME"                       , 114 , 1 , 0)
def_op(loc, "SWAP"                             , 115 , 0 , 0)
def_op(loc, "UNPACK_EX"                        , 116 , VARYING_STACK_INT, VARYING_STACK_INT)
def_op(loc, "UNPACK_SEQUENCE"                  , 117 , 0 , -1)
def_op(loc, "YIELD_VALUE"                      , 118 , 1 , 1)
def_op(loc, "RESUME"                           , 149 , 0 , 0)
def_op(loc, "INSTRUMENTED_RESUME"              , 236 , 1 , 1)
def_op(loc, "INSTRUMENTED_END_FOR"             , 237 , 2 , 1)
def_op(loc, "INSTRUMENTED_END_SEND"            , 238 , 1 , 0)
def_op(loc, "INSTRUMENTED_RETURN_VALUE"        , 239 , 1 , 0)
def_op(loc, "INSTRUMENTED_RETURN_CONST"        , 240 , 1 , 1)
def_op(loc, "INSTRUMENTED_YIELD_VALUE"         , 241 , 1 , 1)
def_op(loc, "INSTRUMENTED_LOAD_SUPER_ATTR"     , 242 , 2 , 0)
def_op(loc, "INSTRUMENTED_FOR_ITER"            , 243 , 1 , 1)
call_op(loc, "INSTRUMENTED_CALL"               , 244 , 1 , 1)
call_op(loc, "INSTRUMENTED_CALL_KW"            , 245 , 2 , 2)
call_op(loc, "INSTRUMENTED_CALL_FUNCTION_EX"   , 246 , 1 , 1)
def_op(loc, "INSTRUMENTED_INSTRUCTION"         , 247 , 1 , 1)
def_op(loc, "INSTRUMENTED_JUMP_FORWARD"        , 248 , 1 , 1)
def_op(loc, "INSTRUMENTED_JUMP_BACKWARD"       , 249 , 1 , 1)
def_op(loc, "INSTRUMENTED_POP_JUMP_IF_FALSE"   , 251 , 1 , 1)
def_op(loc, "INSTRUMENTED_POP_JUMP_IF_NONE"    , 252 , 1 , 1)
def_op(loc, "INSTRUMENTED_POP_JUMP_IF_NOT_NONE", 253 , 1 , 1)
def_op(loc, "JUMP"                             , 256 , 0 , 0)
def_op(loc, "JUMP_NO_INTERRUPT"                , 257 , 0 , 0)
def_op(loc, "LOAD_CLOSURE"                     , 258 , 0 , 1)
def_op(loc, "LOAD_METHOD"                      , 259 , 0 , 1)
def_op(loc, "LOAD_SUPER_METHOD"                , 260 , 1 , 0)
def_op(loc, "LOAD_ZERO_SUPER_ATTR"             , 261 , 1 , 0)
def_op(loc, "LOAD_ZERO_SUPER_METHOD"           , 262 , 1 , 0)
def_op(loc, "POP_BLOCK"                        , 263 , 1 , 1)
def_op(loc, "SETUP_CLEANUP"                    , 264 , 0 , 2)
def_op(loc, "SETUP_FINALLY"                    , 265 , 0 , 1)
def_op(loc, "SETUP_WITH"                       , 266 , 0 , 1)
def_op(loc, "STORE_FAST_MAYBE_NULL"            , 267 , 1 , 0)

## These are new since 3.12...
#            OP NAME                              OPCODE POP PUSH
# ---------------------------------------------------------------
def_op(loc, "FORMAT_SIMPLE"         , 14 , 1 , 1)
def_op(loc, "TO_BOOL"               , 40 , 0 , 0)
def_op(loc, "CALL_KW"               , 57 , -2, 1)
def_op(loc, "CONVERT_VALUE"         , 60 , 1 , 1)
def_op(loc, "SET_FUNCTION_ATTRIBUTE", 106, 2 , 1)

# fmt: on

### update opinfo tables
# completely redefined tables
loc["hasarg"] = [
    45,
    46,
    47,
    48,
    49,
    50,
    51,
    52,
    53,
    54,
    55,
    56,
    57,
    58,
    59,
    60,
    61,
    62,
    63,
    64,
    65,
    66,
    67,
    68,
    69,
    70,
    71,
    72,
    73,
    74,
    75,
    76,
    77,
    78,
    79,
    80,
    81,
    82,
    83,
    84,
    85,
    86,
    87,
    88,
    89,
    90,
    91,
    92,
    93,
    94,
    95,
    96,
    97,
    98,
    99,
    100,
    101,
    102,
    103,
    104,
    105,
    106,
    107,
    108,
    109,
    110,
    111,
    112,
    113,
    114,
    115,
    116,
    117,
    118,
    149,
    236,
    240,
    241,
    242,
    243,
    244,
    245,
    248,
    249,
    250,
    251,
    252,
    253,
    256,
    257,
    258,
    259,
    260,
    261,
    262,
    264,
    265,
    266,
    267,
]
loc["hascompare"] = [58]
loc["hasconst"] = [83, 103, 240]
loc["haslocal"] = [65, 85, 86, 87, 88, 110, 111, 112, 258, 267]
loc["hasname"] = [
    63,
    66,
    67,
    74,
    75,
    82,
    90,
    91,
    92,
    93,
    108,
    113,
    114,
    259,
    260,
    261,
    262,
]
loc["hasfree"] = [64, 84, 89, 94, 109]
# add new table "hasjump"
loc.update({"hasjump": [72, 77, 78, 79, 97, 98, 99, 100, 104, 256, 257]})
loc["hasjrel"] = loc["hasjump"]


def extended_format_CALL(opc, instructions) -> Tuple[str, Optional[int]]:
    """call_method should be a "CALL_METHOD" instruction. Look in
    `instructions` to see if we can find a method name.  If not we'll
    return None.

    """
    # From opcode description: arg_count indicates the total number of
    # positional and keyword arguments.

    call_method_inst = instructions[0]
    arg_count = call_method_inst.argval
    s = ""

    arglist, arg_count, first_arg = get_arglist(instructions, 0, arg_count)
    if arglist is None:
        arglist = []

    if first_arg is None or first_arg >= len(instructions) - 1:
        return NULL_EXTENDED_OP

    fn_inst = instructions[first_arg + 1]
    if fn_inst.opcode in opc.operator_set and arglist is not None:
        start_offset = fn_inst.offset

    if fn_inst.opname == "PUSH_NULL":
        fn_inst = instructions[first_arg + 2]
        if fn_inst.opname in ("LOAD_NAME",):
            start_offset = fn_inst.offset
            fn_name = fn_inst.tos_str if fn_inst.tos_str else fn_inst.argrepr
            arglist.reverse()
            s = f'{fn_name}({", ".join(arglist)})'
            return s, start_offset

    return NULL_EXTENDED_OP


### update formatting
opcode_arg_fmt = opcode_arg_fmt313 = {
    **opcode_312.opcode_arg_fmt312,
    **{"CALL": format_CALL_METHOD},
}

opcode_extended_fmt = opcode_extended_fmt313 = {
    **opcode_312.opcode_extended_fmt312,
    **{"CALL": extended_format_CALL},
}

for fmt_table in (opcode_arg_fmt313, opcode_extended_fmt313):
    fmt_table.pop("MAKE_FUNCTION")  # MAKE_FUNCTION formatting not in 3.13


# update any calls to findlinestarts to include the version tuple
def findlinestarts_313(code, dup_lines: bool = False):
    lastline = False  # None is a valid line number
    for start, _, line in code.co_lines():
        if line is not lastline:
            lastline = line
            yield start, line


findlinestarts = findlinestarts_313

update_pj3(globals(), loc)
finalize_opcodes(loc)



# (C) Copyright 2020-2021, 2023-2025 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import types
from collections import namedtuple
from typing import Any, Dict, Optional, Set, Tuple, Union

from xdis.codetype.base import CodeBase
from xdis.codetype.code13 import Code13
from xdis.codetype.code15 import Code15
from xdis.codetype.code20 import Code2
from xdis.codetype.code30 import Code3
from xdis.codetype.code38 import Code38
from xdis.codetype.code38graal import Code38Graal
from xdis.codetype.code310 import Code310
from xdis.codetype.code310graal import Code310Graal
from xdis.codetype.code311 import Code311, Code311FieldNames
from xdis.codetype.code311graal import Code311Graal
from xdis.version_info import IS_GRAAL, IS_PYPY, PYTHON_VERSION_TRIPLE

__docformat__ = "restructuredtext"


def codeType2Portable(code, version_triple=PYTHON_VERSION_TRIPLE, is_graal: bool=False, other_fields: dict={}):
    """Converts a native types.CodeType code object into a
    corresponding more flexible xdis Code type.
    """
    if isinstance(code, CodeBase):
        return code
    if not (isinstance(code, types.CodeType) or isinstance(code, CodeTypeUnion)):
        raise TypeError(
            f"parameter expected to be a types.CodeType type; is {type(code)} instead"
        )
    line_table_field = (
        "co_lnotab"
        if PYTHON_VERSION_TRIPLE < (3, 11) and hasattr(code, "co_lnotab")
        else "co_linetable"
    )
    line_table = getattr(code, line_table_field)
    if version_triple >= (3, 0):
        if version_triple < (3, 8):
            return Code3(
                code.co_argcount,
                code.co_kwonlyargcount,
                code.co_nlocals,
                code.co_stacksize,
                code.co_flags,
                code.co_code,
                code.co_consts,
                code.co_names,
                code.co_varnames,
                code.co_filename,
                code.co_name,
                code.co_firstlineno,
                line_table,
                code.co_freevars,
                code.co_cellvars,
                # THINK ABOUT: If collection_order isn't defined, i.e. native code
                # type, should we try to extract it?
                code.collection_order if hasattr(code, "collection_order") else {},
                code.reference_objects if hasattr(code, "reference_objects") else set(),
                version_triple=version_triple,
            )
        elif version_triple < (3, 10):
            if is_graal:
                other_fields = {}
                # other_fields["condition_profileCount"]=code.condition_profileCount if hasattr(code, "condition_profileCount") else -1,
                # other_fields["endColumn"]=code.endColumn if hasattr(code, "endColumn") else -1,
                # other_fields["endLine"]=code.endLine if hasattr(code, "endLine") else -1,
                # other_fields["exception_handler_ranges"]=code.exception_handler_ranges if hasattr(code, "exception_handler_ranges") else tuple(),
                # other_fields["generalizeInputsMap"]=code.generalizeInputsMap if hasattr(code, "generalizeInputsMap") else {},
                # other_fields["generalizeVarsMap"]=code.generalizeVarsMap if hasattr(code, "generalizeVarsMap") else {},
                # other_fields["outputCanQuicken"]=code.outputCanQuicken if hasattr(code, "outputCanQuicken") else b"",
                # other_fields["primitiveConstants"]=code.primitiveConstants if hasattr(code, "primitiveConstants") else tuple(),
                # other_fields["srcOffsetTable"]=code.srcOffsetTable if hasattr(code, "srcOffsetTable") else b"",
                # other_fields["startColumn"]=code.startColumn if hasattr(code, "startColumn") else -1,
                # other_fields["startLine"]=code.startLine if hasattr(code, "startLine") else -1,
                # other_fields["variableShouldUnbox"]=code.variableShouldUnbox if hasattr(code, "variableShouldUnbox") else b"",
                return Code38Graal(
                    co_argcount=code.co_argcount,
                    co_posonlyargcount=code.co_posonlyargcount,
                    co_kwonlyargcount=code.co_kwonlyargcount,
                    co_nlocals=code.co_nlocals,
                    co_stacksize=code.co_stacksize,
                    co_flags=code.co_flags,
                    co_consts=code.co_consts,
                    co_code=code.co_code,
                    co_names=code.co_names,
                    co_varnames=code.co_varnames,
                    co_freevars=code.co_freevars,
                    co_cellvars=code.co_cellvars,
                    co_filename=code.co_filename,
                    co_name=code.co_name,
                    co_firstlineno=code.co_firstlineno,
                    co_lnotab=line_table,
                    version_triple=version_triple,
                    other_fields=other_fields,
                )
            else:
                return Code38(
                    co_argcount=code.co_argcount,
                    co_posonlyargcount=code.co_posonlyargcount,
                    co_kwonlyargcount=code.co_kwonlyargcount,
                    co_nlocals=code.co_nlocals,
                    co_stacksize=code.co_stacksize,
                    co_flags=code.co_flags,
                    co_code=code.co_code,
                    co_consts=code.co_consts,
                    co_names=code.co_names,
                    co_varnames=code.co_varnames,
                    co_freevars=code.co_freevars,
                    co_cellvars=code.co_cellvars,
                    co_filename=code.co_filename,
                    co_name=code.co_name,
                    co_firstlineno=code.co_firstlineno,
                    co_lnotab=line_table,
                    version_triple=version_triple,
                )
        elif version_triple[:2] == (3, 10) or IS_PYPY and version_triple[:2] == (3, 11):
            if is_graal:
                return Code310Graal(
                    co_argcount=code.co_argcount,
                    co_posonlyargcount=code.co_posonlyargcount,
                    co_kwonlyargcount=code.co_kwonlyargcount,
                    co_nlocals=code.co_nlocals,
                    co_stacksize=code.co_stacksize,
                    co_flags=code.co_flags,
                    co_code=code.co_code,
                    co_consts=code.co_consts,
                    co_names=code.co_names,
                    co_varnames=code.co_varnames,
                    co_freevars=code.co_freevars,
                    co_cellvars=code.co_cellvars,
                    co_filename=code.co_filename,
                    co_name=code.co_name,
                    co_firstlineno=code.co_firstlineno,
                    co_linetable=line_table,
                    version_triple=version_triple,
                    other_fields=other_fields,
                )
            else:
                return Code310(
                    co_argcount=code.co_argcount,
                    co_posonlyargcount=code.co_posonlyargcount,
                    co_kwonlyargcount=code.co_kwonlyargcount,
                    co_nlocals=code.co_nlocals,
                    co_stacksize=code.co_stacksize,
                    co_flags=code.co_flags,
                    co_code=code.co_code,
                    co_consts=code.co_consts,
                    co_names=code.co_names,
                    co_varnames=code.co_varnames,
                    co_freevars=code.co_freevars,
                    co_cellvars=code.co_cellvars,
                    co_filename=code.co_filename,
                    co_name=code.co_name,
                    co_firstlineno=code.co_firstlineno,
                    co_linetable=line_table,
                    version_triple=version_triple,
                )
        elif version_triple[:2] >= (3, 11):
            if is_graal:

                return Code311Graal(
                    co_argcount=code.co_argcount,
                    co_posonlyargcount=code.co_posonlyargcount,
                    co_kwonlyargcount=code.co_kwonlyargcount,
                    co_nlocals=code.co_nlocals,
                    co_stacksize=code.co_stacksize,
                    co_flags=code.co_flags,
                    co_consts=code.co_consts,
                    co_code=code.co_code,
                    co_names=code.co_names,
                    co_varnames=code.co_varnames,
                    co_freevars=code.co_freevars,
                    co_cellvars=code.co_cellvars,
                    co_filename=code.co_filename,
                    co_name=code.co_name,
                    co_qualname=code.co_qualname,
                    co_firstlineno=code.co_firstlineno,
                    co_linetable=line_table,
                    co_exceptiontable=code.co_exceptiontable,
                    version_triple=version_triple,
                    other_fields=other_fields,
                )
            else:
                return Code311(
                    co_argcount=code.co_argcount,
                    co_posonlyargcount=code.co_posonlyargcount,
                    co_kwonlyargcount=code.co_kwonlyargcount,
                    co_nlocals=code.co_nlocals,
                    co_stacksize=code.co_stacksize,
                    co_flags=code.co_flags,
                    co_consts=code.co_consts,
                    co_code=code.co_code,
                    co_names=code.co_names,
                    co_varnames=code.co_varnames,
                    co_freevars=code.co_freevars,
                    co_cellvars=code.co_cellvars,
                    co_filename=code.co_filename,
                    co_name=code.co_name,
                    co_qualname=code.co_qualname,
                    co_firstlineno=code.co_firstlineno,
                    co_linetable=line_table,
                    co_exceptiontable=code.co_exceptiontable,
                    version_triple=version_triple,
                )
    elif version_triple > (2, 0):
        # 2.0 .. 2.7
        return Code2(
            co_argcount=code.co_argcount,
            co_nlocals=code.co_nlocals,
            co_stacksize=code.co_stacksize,
            co_flags=code.co_flags,
            co_code=code.co_code,
            co_consts=code.co_consts,
            co_names=code.co_names,
            co_varnames=code.co_varnames,
            co_filename=code.co_filename,
            co_name=code.co_name,
            co_firstlineno=code.co_firstlineno,
            co_lnotab=line_table,
            co_freevars=code.co_freevars,  # not in 1.x
            co_cellvars=code.co_cellvars,  # not in 1.x
            # THINK ABOUT: If collection_order isn't defined, i.e. native code
            # type, should we try to extract it?
            collection_order=(
                code.collection_order if hasattr(code, "collection_order") else {}
            ),
            reference_objects=(
                code.reference_objects if hasattr(code, "reference_objects") else set()
            ),
            version_triple=version_triple,
        )
    else:
        # 1.0 .. 1.5
        if version_triple < (1, 5):
            # 1.0 .. 1.3
            return Code13(
                code.co_argcount,
                code.co_nlocals,
                code.co_flags,
                code.co_code,
                code.co_consts,
                code.co_names,
                code.co_varnames,
                code.co_filename,
                code.co_name,
            )
        else:
            return Code15(
                code.co_argcount,
                code.co_nlocals,
                code.co_stacksize,  # not in 1.0..1.4
                code.co_flags,
                code.co_code,
                code.co_consts,
                code.co_names,
                code.co_varnames,
                code.co_filename,
                code.co_name,
                code.co_firstlineno,  # Not in 1.0..1.4
                line_table,  # Not in 1.0..1.4
            )


def portableCodeType(version_triple=PYTHON_VERSION_TRIPLE, is_graal=IS_GRAAL):
    """
    Return the portable CodeType version for the supplied Python release version.
    `version` is a floating-point number, like 2.7, or 3.9. If no version
    number is supplied we'll use the current interpreter version.
    """
    if version_triple >= (3, 0):
        if version_triple < (3, 8):
            # 3.0 .. 3.7
            return Code3
        elif version_triple < (3, 10):
            # 3.8 ... 3.9
            return Code38
        elif version_triple[:2] == (3, 10) or IS_PYPY and version_triple[:2] == (3, 11):
            # 3.10
            if is_graal:
                return Code310Graal
            else:
                return Code310
        elif version_triple[:2] >= (3, 11):
            # 3.11 ...
            if is_graal:
                return Code311Graal
            else:
                return Code311
    elif version_triple > (2, 0):
        # 2.0 .. 2.7
        return Code2
    else:
        # 1.0 .. 1.5
        if version_triple <= (1, 3):
            return Code13
        else:
            return Code15


# In contrast to Code3, Code2, etc. you can use CodeTypeUnint for building
# an incomplete code type, which might be converted to another code type
# later.
CodeTypeUnionFields = tuple(
    Code311FieldNames.split()
    + ["collection_order", "reference_objects", "version_triple", "other_fields"]
)
CodeTypeUnion = namedtuple("CodeTypeUnion", CodeTypeUnionFields)


# Note: default values of `None` indicate a required parameter.
# default values of -1, (None,) or "" indicate an unsupplied parameter.
def to_portable(
    co_argcount: int,
    co_posonlyargcount: Optional[int] = -1,  # 3.8 .. 3.10
    co_kwonlyargcount: Optional[int] = -1,  # 3.0+
    co_nlocals: int = 0,
    co_stacksize: Optional[int] = -1,  # 1.5+
    co_flags: int = 0,
    co_code: Union[str, bytes] = "",  # 3.0+ this type changes from <str> to <bytes>
    co_consts: tuple[str, ...] = tuple(),
    co_names: tuple[str, ...] = tuple(),
    co_varnames: tuple[str, ...] = tuple(),
    co_filename: str = "??",
    co_name: str = "??",
    co_qualname: str = "??",
    co_firstlineno: int = -1,
    co_lnotab: str = "",  # 1.5+; 3.0+ this type changes from <str> to <bytes>
    # In 3.11 it is different
    co_freevars: tuple[str, ...] = tuple(),  # 2.0+
    co_cellvars: tuple[str, ...] = tuple(),  # 2.0+
    co_exceptiontable=None,  # 3.11+
    version_triple=PYTHON_VERSION_TRIPLE,
    collection_order: Dict[Union[set, frozenset, dict], Tuple[Any]] = {},
    reference_objects: Set[Any] = set(),
    other_fields: Dict[str, Any] = {},
):
    is_graal = "srcOffsetTable" in other_fields
    code = CodeTypeUnion(
        co_argcount=co_argcount,
        co_posonlyargcount=co_posonlyargcount,
        co_kwonlyargcount=co_kwonlyargcount,
        co_nlocals=co_nlocals,
        co_stacksize=co_stacksize,
        co_flags=co_flags,
        co_code=co_code,
        co_consts=co_consts,
        co_names=co_names,
        co_varnames=co_varnames,
        co_filename=co_filename,
        co_name=co_name,
        co_qualname=co_name if co_qualname is None else co_qualname,
        co_firstlineno=co_firstlineno,
        co_linetable=co_lnotab,
        co_lnotab=co_lnotab,
        co_freevars=co_freevars,
        co_cellvars=co_cellvars,
        co_exceptiontable=co_exceptiontable,
        collection_order=collection_order,
        reference_objects=reference_objects,
        version_triple=version_triple,
        other_fields=other_fields,
    )
    return codeType2Portable(code, version_triple, is_graal=is_graal, other_fields=other_fields)


if __name__ == "__main__":
    x = codeType2Portable(to_portable.__code__)
    print(x)
